# Copyright 2015 BMW Car IT GmbH
# Copyright 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import threading
import time
import textwrap
from unittest import mock
import yaml

import zuul.model
import tests.base
from tests.base import (
    AnsibleZuulTestCase,
    BaseTestCase,
    ZuulTestCase,
    gerrit_config,
    iterate_timeout,
    okay_tracebacks,
    simple_layout,
    skipIfMultiScheduler,
)
from zuul.lib import strings
from zuul.driver.gerrit import GerritDriver
from zuul.driver.gerrit.gerritconnection import (
    ChangeNetworkConflict,
    GerritConnection,
    PeekQueue,
)

import git
import paramiko
import testtools

FIXTURE_DIR = os.path.join(tests.base.FIXTURE_DIR, 'gerrit')


def read_fixture(file):
    with open('%s/%s' % (FIXTURE_DIR, file), 'r') as fixturefile:
        lines = fixturefile.readlines()
        command = lines[0].replace('\n', '')
        value = ''.join(lines[1:])
        return command, value


def read_fixtures(files):
    calls = []
    values = []
    for fixture_file in files:
        command, value = read_fixture(fixture_file)
        calls.append(mock.call(command))
        values.append([value, ''])
    return calls, values


class TestGerrit(BaseTestCase):

    @mock.patch('zuul.driver.gerrit.gerritconnection.GerritConnection._ssh')
    def run_query(self, files, expected_patches, _ssh_mock):
        gerrit_config = {
            'user': 'gerrit',
            'server': 'localhost',
        }
        driver = GerritDriver()
        gerrit = GerritConnection(driver, 'review_gerrit', gerrit_config)

        calls, values = read_fixtures(files)
        _ssh_mock.side_effect = values

        result = gerrit.simpleQuery('project:zuul/zuul')

        _ssh_mock.assert_has_calls(calls)
        self.assertEqual(len(calls), _ssh_mock.call_count,
                         '_ssh should be called %d times' % len(calls))
        self.assertIsNotNone(result, 'Result is not none')
        self.assertEqual(len(result), expected_patches,
                         'There must be %d patches.' % expected_patches)

    def test_simple_query_pagination_new(self):
        files = ['simple_query_pagination_new_1',
                 'simple_query_pagination_new_2']
        expected_patches = 5
        self.run_query(files, expected_patches)

    def test_simple_query_pagination_old(self):
        files = ['simple_query_pagination_old_1',
                 'simple_query_pagination_old_2',
                 'simple_query_pagination_old_3']
        expected_patches = 5
        self.run_query(files, expected_patches)

    def test_ref_name_check_rules(self):
        # See man git-check-ref-format for the rules referenced here
        test_strings = [
            ('refs/heads/normal', True),
            ('refs/heads/.bad', False),  # rule 1
            ('refs/heads/bad.lock', False),  # rule 1
            ('refs/heads/good.locked', True),
            ('refs/heads/go.od', True),
            ('refs/heads//bad', False),  # rule 6
            ('refs/heads/b?d', False),  # rule 5
            ('refs/heads/b[d', False),  # rule 5
            ('refs/heads/b..ad', False),  # rule 3
            ('bad', False),  # rule 2
            ('refs/heads/\nbad', False),  # rule 4
            ('/refs/heads/bad', False),  # rule 6
            ('refs/heads/bad/', False),  # rule 6
            ('refs/heads/bad.', False),  # rule 7
            ('.refs/heads/bad', False),  # rule 1
            ('refs/he@{ads/bad', False),  # rule 8
            ('@', False),  # rule 9
            ('refs\\heads/bad', False)  # rule 10
        ]

        for ref, accepted in test_strings:
            self.assertEqual(
                accepted,
                GerritConnection._checkRefFormat(ref),
                ref + ' shall be ' + ('accepted' if accepted else 'rejected'))

    def test_getGitURL(self):
        gerrit_config = {
            'user': 'gerrit',
            'server': 'localhost',
            'password': '1/badpassword',
        }
        # The 1/ in the password ensures we test the url encoding
        # path; this is the format of password we get from
        # googlesource.com.
        driver = GerritDriver()
        gerrit = GerritConnection(driver, 'review_gerrit', gerrit_config)
        project = gerrit.source.getProject('org/project')
        url = gerrit.source.getGitUrl(project)
        self.assertEqual(
            'https://gerrit:1%2Fbadpassword@localhost/a/org/project',
            url)

    def test_git_over_ssh_getGitURL(self):
        gerrit_config = {
            'user': 'gerrit',
            'server': 'localhost',
            'password': '1/badpassword',
            'git_over_ssh': 'true',
        }
        # The 1/ in the password ensures we test the url encoding
        # path; this is the format of password we get from
        # googlesource.com.
        driver = GerritDriver()
        gerrit = GerritConnection(driver, 'review_gerrit', gerrit_config)
        project = gerrit.source.getProject('org/project')
        url = gerrit.source.getGitUrl(project)
        self.assertEqual(
            'ssh://gerrit@localhost:29418/org/project',
            url)

    def test_ssh_server_getGitURL(self):
        gerrit_config = {
            'user': 'gerrit',
            'server': 'otherserver',
            'password': '1/badpassword',
            'ssh_server': 'localhost',
            'git_over_ssh': 'true',
        }
        # The 1/ in the password ensures we test the url encoding
        # path; this is the format of password we get from
        # googlesource.com.
        driver = GerritDriver()
        gerrit = GerritConnection(driver, 'review_gerrit', gerrit_config)
        project = gerrit.source.getProject('org/project')
        url = gerrit.source.getGitUrl(project)
        self.assertEqual(
            'ssh://gerrit@localhost:29418/org/project',
            url)

    def test_ssh_exec_timeout(self):
        with mock.patch('paramiko.SSHClient') as mock_ssh_client:
            mock_client = mock.Mock()
            mock_ssh_client.return_value = mock_client
            stdin = mock.Mock()
            stdout = mock.Mock()
            stdout.channel = mock.Mock()
            stdout.channel.recv_exit_status.return_value = 0
            stderr = mock.Mock()
            mock_client.exec_command.return_value = (stdin, stdout, stderr)
            gerrit_config = {
                'user': 'gerrit',
                'server': 'localhost',
            }
            driver = GerritDriver()
            gerrit = GerritConnection(driver, 'review_gerrit', gerrit_config)
            gerrit._ssh("echo hi")
            expected = [
                mock.call('echo hi', timeout=30),
            ]
            self.assertEqual(expected, mock_client.exec_command.mock_calls)

        with mock.patch('paramiko.SSHClient') as mock_ssh_client:
            mock_client = mock.Mock()
            mock_ssh_client.return_value = mock_client
            mock_client.exec_command.side_effect =\
                paramiko.buffered_pipe.PipeTimeout()

            gerrit_config = {
                'user': 'gerrit',
                'server': 'localhost',
            }
            driver = GerritDriver()
            gerrit = GerritConnection(driver, 'review_gerrit', gerrit_config)
            with testtools.ExpectedException(
                    paramiko.buffered_pipe.PipeTimeout):
                gerrit._ssh("echo hi")
            expected = [
                mock.call('echo hi', timeout=30),
                mock.call('echo hi', timeout=30),
            ]
            self.assertEqual(expected, mock_client.exec_command.mock_calls)


class TestGerritWeb(ZuulTestCase):
    config_file = 'zuul-gerrit-web.conf'
    tenant_config_file = 'config/single-tenant/main.yaml'

    def test_jobs_executed(self):
        "Test that jobs are executed and a change is merged"
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        A.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()
        self.assertEqual(self.getJobFromHistory('project-merge').result,
                         'SUCCESS')
        self.assertEqual(self.getJobFromHistory('project-test1').result,
                         'SUCCESS')
        self.assertEqual(self.getJobFromHistory('project-test2').result,
                         'SUCCESS')
        self.assertEqual(A.data['status'], 'MERGED')
        self.assertEqual(A.reported, 2)
        self.assertEqual(self.getJobFromHistory('project-test1').node,
                         'label1')
        self.assertEqual(self.getJobFromHistory('project-test2').node,
                         'label1')

    def test_dynamic_line_comment(self):
        in_repo_conf = textwrap.dedent(
            """
            - job:
                name: garbage-job
                garbage: True
            """)

        file_dict = {'.zuul.yaml': in_repo_conf}
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A',
                                           files=file_dict)
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.assertEqual(A.patchsets[0]['approvals'][0]['value'], "-1")
        self.assertEqual(A.patchsets[0]['approvals'][0]['__tag'],
                         "autogenerated:zuul:check")
        self.assertIn('Zuul encountered a syntax error',
                      A.messages[0])
        comments = sorted(A.comments, key=lambda x: x['line'])
        self.assertEqual(comments[0],
                         {'file': '.zuul.yaml',
                          'line': 4,
                          'message': "extra keys not allowed @ "
                                     "data['garbage']",
                          'range': {'end_character': 0,
                                    'end_line': 4,
                                    'start_character': 2,
                                    'start_line': 2},
                          'reviewer': {'email': 'zuul@example.com',
                                       'name': 'Zuul',
                                       'username': 'jenkins'}}
        )

    def test_message_too_long(self):
        in_repo_conf = textwrap.dedent(
            """
            - job:
                name: garbage-job
                description: %s
                garbage: True
            """
        ) % ('x' * 16384)

        file_dict = {'.zuul.yaml': in_repo_conf}
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A',
                                           files=file_dict)
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.assertEqual(A.patchsets[0]['approvals'][0]['value'], "-1")
        self.assertEqual(A.patchsets[0]['approvals'][0]['__tag'],
                         "autogenerated:zuul:check")
        self.assertIn('... (truncated)',
                      A.messages[0])

    def test_message_too_long_failed_job(self):
        # Test that if we have a failed job in a long message, we
        # at least include the failed jobs.
        config = []
        joblist = []
        # Use enough jobs with long names to trip the 16k limit
        for x in range(60):
            name = 'job-%s-%s' % ('x' * 200, x)
            job = dict(
                name=name,
                parent='test-job',
            )
            config.append(dict(job=job))
            joblist.append(name)
        config.append(dict(
            project=dict(
                check=dict(
                    jobs=joblist,
                )
            )
        ))
        in_repo_conf = yaml.dump(config)
        file_dict = {'.zuul.yaml': in_repo_conf}
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A',
                                           files=file_dict)
        self.executor_server.failJob('project-test1', A)
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.assertIn('length restrictions', A.messages[0])
        self.assertIn('FAILURE', A.messages[0])
        self.assertNotIn('SUCCESS', A.messages[0])

    def test_dependent_dynamic_line_comment(self):
        in_repo_conf = textwrap.dedent(
            """
            - job:
                name: garbage-job
                garbage: True
            """)

        file_dict = {'.zuul.yaml': in_repo_conf}
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A',
                                           files=file_dict)
        B = self.fake_gerrit.addFakeChange('org/project1', 'master', 'B')
        B.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            B.subject, A.data['id'])

        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.assertEqual(B.patchsets[0]['approvals'][0]['value'], "-1")
        self.assertIn('This change depends on a change '
                      'with an invalid configuration',
                      B.messages[0])
        self.assertEqual(B.comments, [])

    @simple_layout('layouts/single-file-matcher.yaml')
    def test_single_file(self):
        # HTTP requests don't return a commit_msg entry in the files
        # list, but the rest of zuul always expects one.  This test
        # returns a single file to exercise the single-file code path
        # in the files matcher.
        files = {'README': 'please!\n'}

        change = self.fake_gerrit.addFakeChange('org/project',
                                                'master',
                                                'test irrelevant-files',
                                                files=files)
        self.fake_gerrit.addEvent(change.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        tested_change_ids = [x.changes[0] for x in self.history
                             if x.name == 'project-test-irrelevant-files']

        self.assertEqual([], tested_change_ids)

    def test_recheck(self):
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(3, len(self.history))

        self.fake_gerrit.addEvent(A.getChangeCommentEvent(1))
        self.waitUntilSettled()
        self.assertEqual(3, len(self.history))

        self.fake_gerrit.addEvent(A.getChangeCommentEvent(1,
                                  'recheck'))
        self.waitUntilSettled()
        self.assertEqual(6, len(self.history))

        self.fake_gerrit.addEvent(A.getChangeCommentEvent(1,
                                  patchsetcomment='recheck'))
        self.waitUntilSettled()
        self.assertEqual(9, len(self.history))

        self.fake_gerrit.addEvent(A.getChangeCommentEvent(1,
                                  patchsetcomment='do not recheck'))
        self.waitUntilSettled()
        self.assertEqual(9, len(self.history))

    def test_submitted_together_git(self):
        # This tests that the circular dependency handling for submit
        # whole topic doesn't activate for changes which are only in a
        # git dependency.
        A = self.fake_gerrit.addFakeChange('org/project1', "master", "A")
        B = self.fake_gerrit.addFakeChange('org/project1', "master", "B")
        C = self.fake_gerrit.addFakeChange('org/project1', "master", "C")
        D = self.fake_gerrit.addFakeChange('org/project1', "master", "D")
        E = self.fake_gerrit.addFakeChange('org/project1', "master", "E")
        F = self.fake_gerrit.addFakeChange('org/project1', "master", "F")
        G = self.fake_gerrit.addFakeChange('org/project1', "master", "G")
        G.setDependsOn(F, 1)
        F.setDependsOn(E, 1)
        E.setDependsOn(D, 1)
        D.setDependsOn(C, 1)
        C.setDependsOn(B, 1)
        B.setDependsOn(A, 1)

        self.fake_gerrit.addEvent(C.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.assertEqual(len(C.patchsets[-1]["approvals"]), 1)
        self.assertEqual(C.patchsets[-1]["approvals"][0]["type"], "Verified")
        self.assertEqual(C.patchsets[-1]["approvals"][0]["value"], "1")
        self.assertEqual(A.queried, 1)
        self.assertEqual(B.queried, 1)
        self.assertEqual(C.queried, 1)
        self.assertEqual(D.queried, 1)
        self.assertEqual(E.queried, 1)
        self.assertEqual(F.queried, 1)
        self.assertEqual(G.queried, 1)
        self.assertHistory([
            dict(name="project-merge", result="SUCCESS",
                 changes="1,1 2,1 3,1"),
            dict(name="project-test1", result="SUCCESS",
                 changes="1,1 2,1 3,1"),
            dict(name="project-test2", result="SUCCESS",
                 changes="1,1 2,1 3,1"),
            dict(name="project1-project2-integration", result="SUCCESS",
                 changes="1,1 2,1 3,1"),
        ], ordered=False)

    def test_submit_failure(self):
        # Test that we log the reason for a submit failure (403 error)
        self.fake_gerrit._fake_submit_permission = False
        A = self.fake_gerrit.addFakeChange('org/project1', "master", "A")
        A.addApproval('Code-Review', 2)
        with self.assertLogs('zuul.test.FakeGerritConnection', level='INFO'
                             ) as full_logs:
            self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
            self.waitUntilSettled()
            self.log.debug("Full logs:")
            for x in full_logs.output:
                self.log.debug(x)
            self.assertRegexInList(
                r'Error submitting data to gerrit on attempt 3: '
                'Received response 403: submit not permitted',
                full_logs.output)

        self.assertEqual(A.data['status'], 'NEW')

    @okay_tracebacks('_handleEvent')
    def test_event_failure(self):
        # Test that a failure to process a Gerrit event is handled
        # correctly
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        event = A.getPatchsetCreatedEvent(1)
        self.fake_gerrit._fake_return_api_error = True
        self.fake_gerrit.addEvent(event)
        # If the event is not processed, then we will never settle and
        # the test will time-out.
        self.waitUntilSettled()

    def test_get_project_branch_sha(self):
        # Exercise this method since it's only called from timer
        # triggers
        source = self.fake_gerrit.source
        project = source.getProject('org/project')
        self.assertIsNotNone(source.getProjectBranchSha(project, 'master'))

    @simple_layout('layouts/gerrit-pipeline-trigger-debug.yaml')
    def test_gerrit_pipeline_trigger_debug(self):
        # Test that we get debug info for a pipeline debug trigger
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'NEW')
        self.assertEqual(A.reported, 1)
        self.assertIn('Debug information:',
                      A.messages[0])

        # Test that we get debug info for a pipeline debug trigger
        # when a change depends on an abandoned change
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        # We check approvals before state, so set those
        B.addApproval('Code-Review', 2)
        B.addApproval('Approved', 1)
        B.setAbandoned()
        A.setDependsOn(B, 1)
        A.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'NEW')
        self.assertEqual(A.reported, 2)
        # Full message: Change 2,1 in project org/project is needed
        # but can not be merged due to: change is closed
        # Meaning we can not enqueue A because it depends on B which
        # is abandoned.
        self.assertIn('change is closed',
                      A.messages[1])

        # Test that we get debug info for a pipeline debug trigger
        # when a change is missing a pipeline requirement
        self.fake_gerrit.addEvent(B.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertEqual(B.data['status'], 'ABANDONED')
        self.assertEqual(B.reported, 1)
        self.assertIn('does not match pipeline requirement',
                      B.messages[0])

        # Test that we get debug info for a pipeline debug trigger
        # when a change is missing a merge requirement
        C = self.fake_gerrit.addFakeChange('org/project', 'master', 'C')
        self.fake_gerrit.addEvent(C.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertEqual(C.data['status'], 'NEW')
        self.assertEqual(C.reported, 1)
        self.assertIn('can not be merged due to: missing labels:',
                      C.messages[0])

        # Test that we get debug info for a pipeline debug trigger
        # when a change depends on a change missing a merge requirement
        D = self.fake_gerrit.addFakeChange('org/project', 'master', 'D')
        D.setDependsOn(C, 1)
        D.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(D.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertEqual(D.data['status'], 'NEW')
        self.assertEqual(D.reported, 1)
        self.assertIn(
            'is needed but can not be merged due to: missing labels:',
            D.messages[0])


class TestFileComments(AnsibleZuulTestCase):
    config_file = 'zuul-gerrit-web.conf'
    tenant_config_file = 'config/gerrit-file-comments/main.yaml'

    def test_file_comments(self):
        A = self.fake_gerrit.addFakeChange(
            'org/project', 'master', 'A',
            files={'path/to/file.py': 'test1',
                   'otherfile.txt': 'test2',
                   })
        A.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(self.getJobFromHistory('file-comments').result,
                         'SUCCESS')
        self.assertEqual(self.getJobFromHistory('file-comments-error').result,
                         'SUCCESS')
        self.assertEqual(len(A.comments), 7)
        comments = sorted(A.comments, key=lambda x: (x['file'], x['line']))
        self.assertEqual(
            comments[0],
            {
                'file': '/COMMIT_MSG',
                'line': 1,
                'message': 'commit message comment',
                'reviewer': {
                    'email': 'zuul@example.com',
                    'name': 'Zuul',
                    'username': 'jenkins'
                },
            },
        )

        self.assertEqual(
            comments[1],
            {
                'file': 'otherfile.txt',
                'line': 21,
                'message': 'This is a much longer message.\n\n'
                'With multiple paragraphs.\n',
                'reviewer': {
                    'email': 'zuul@example.com',
                    'name': 'Zuul',
                    'username': 'jenkins'
                },
            },
        )

        self.assertEqual(
            comments[2],
            {
                "file": "path/to/file.py",
                "line": 2,
                "message": "levels are ignored by gerrit",
                "reviewer": {
                    "email": "zuul@example.com",
                    "name": "Zuul",
                    "username": "jenkins",
                },
            },
        )

        self.assertEqual(
            comments[3],
            {
                "file": "path/to/file.py",
                "line": 21,
                "message": (
                    "A second zuul return value using the same file should not"
                    "\noverride the first result, but both should be merged.\n"
                ),
                "reviewer": {
                    "email": "zuul@example.com",
                    "name": "Zuul",
                    "username": "jenkins",
                },
            },
        )

        self.assertEqual(
            comments[4],
            {
                'file': 'path/to/file.py',
                'line': 42,
                'message': 'line too long',
                'reviewer': {
                    'email': 'zuul@example.com',
                    'name': 'Zuul',
                    'username': 'jenkins'
                },
            },
        )

        self.assertEqual(
            comments[5],
            {
                "file": "path/to/file.py",
                "line": 42,
                "message": (
                    "A second comment applied to the same line in the same "
                    "file\nshould also be added to the result.\n"
                ),
                "reviewer": {
                    "email": "zuul@example.com",
                    "name": "Zuul",
                    "username": "jenkins",
                }
            }
        )

        self.assertEqual(comments[6],
                         {'file': 'path/to/file.py',
                          'line': 82,
                          'message': 'line too short',
                          'reviewer': {'email': 'zuul@example.com',
                                       'name': 'Zuul',
                                       'username': 'jenkins'}}
        )
        self.assertIn('expected a dictionary', A.messages[0],
                      "A should have a validation error reported")
        self.assertIn('invalid file missingfile.txt', A.messages[0],
                      "A should have file error reported")

    def test_merge_commit_file_comments(self):
        self.create_branch('org/project', 'stable/queens')
        self.fake_gerrit.addEvent(
            self.fake_gerrit.getFakeBranchCreatedEvent(
                'org/project', 'stable/queens'))
        self.waitUntilSettled()

        A = self.fake_gerrit.addFakeChange('org/project', 'stable/queens', 'A',
                                           files={'path/to/file.py': 'test1',
                                                  'otherfile.txt': 'test2',
                                                  })
        A.setMerged()
        self.fake_gerrit.addEvent(A.getChangeMergedEvent())
        self.waitUntilSettled()

        upstream_path = os.path.join(self.upstream_root, 'org/project')
        upstream_repo = git.Repo(upstream_path)
        master_sha = upstream_repo.heads.master.commit.hexsha

        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B',
                                           merge_parents=[
                                               master_sha,
                                               A.patchsets[-1]['revision'],
                                           ],
                                           files={'path/to/file.py': 'test1',
                                                  'otherfile.txt': 'test2',
                                                  })

        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertHistory([
            dict(name='file-comments', result='SUCCESS', changes='2,1'),
            dict(name='file-comments-error', result='SUCCESS', changes='2,1'),
        ], ordered=False)
        self.assertEqual(len(B.comments), 1)
        comments = sorted(B.comments, key=lambda x: (x['file'], x['line']))
        self.assertEqual(
            comments[0],
            {
                'file': '/COMMIT_MSG',
                'line': 1,
                'message': 'commit message comment',
                'reviewer': {
                    'email': 'zuul@example.com',
                    'name': 'Zuul',
                    'username': 'jenkins'
                },
            },
        )
        self.assertIn('invalid file missingfile.txt', B.messages[0],
                      "A should have file error reported")


class TestChecksApi(ZuulTestCase):
    config_file = 'zuul-gerrit-web.conf'

    @simple_layout('layouts/gerrit-checks.yaml')
    def test_check_pipeline(self):
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        A.setDependsOn(B, 1)
        A.setCheck('zuul:check', reset=True)
        self.waitForPoll('gerrit')
        self.waitUntilSettled()

        self.assertEqual(A.checks_history[0]['zuul:check']['state'],
                         'NOT_STARTED')

        self.assertEqual(A.checks_history[1]['zuul:check']['state'],
                         'SCHEDULED')
        self.assertEqual(
            A.checks_history[1]['zuul:check']['url'],
            'http://zuul.example.com/t/tenant-one/status/change/2,1')

        self.assertEqual(A.checks_history[2]['zuul:check']['state'],
                         'RUNNING')
        self.assertEqual(
            A.checks_history[2]['zuul:check']['url'],
            'http://zuul.example.com/t/tenant-one/status/change/2,1')

        self.assertEqual(A.checks_history[3]['zuul:check']['state'],
                         'SUCCESSFUL')
        self.assertTrue(
            A.checks_history[3]['zuul:check']['url'].startswith(
                'http://zuul.example.com/t/tenant-one/buildset/'))

        self.assertEqual(len(A.checks_history), 4)

        self.assertTrue(isinstance(
            A.checks_history[3]['zuul:check']['started'], str))
        self.assertTrue(isinstance(
            A.checks_history[3]['zuul:check']['finished'], str))
        self.assertTrue(
            A.checks_history[3]['zuul:check']['finished'] >
            A.checks_history[3]['zuul:check']['started'])
        self.assertEqual(A.checks_history[3]['zuul:check']['message'],
                         'Change passed all voting jobs')
        self.assertHistory([
            dict(name='test-job', result='SUCCESS', changes='1,1 2,1')])
        self.assertEqual(A.reported, 0, "no messages should be reported")
        self.assertEqual(A.messages, [], "no messages should be reported")
        # Make sure B was never updated
        self.assertEqual(len(B.checks_history), 0)

    @simple_layout('layouts/gerrit-checks.yaml')
    def test_gate_pipeline(self):
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        A.addApproval('Code-Review', 2)
        A.addApproval('Approved', 1)
        A.setCheck('zuul:gate', reset=True)
        self.waitForPoll('gerrit')
        self.waitUntilSettled()

        self.assertEqual(A.checks_history[0]['zuul:gate']['state'],
                         'NOT_STARTED')
        self.assertEqual(A.checks_history[1]['zuul:gate']['state'],
                         'SCHEDULED')
        self.assertEqual(A.checks_history[2]['zuul:gate']['state'],
                         'RUNNING')
        self.assertEqual(A.checks_history[3]['zuul:gate']['state'],
                         'SUCCESSFUL')
        self.assertEqual(len(A.checks_history), 4)
        self.assertHistory([
            dict(name='test-job', result='SUCCESS', changes='1,1')])
        self.assertEqual(A.data['status'], 'MERGED')
        self.assertEqual(A.reported, 2,
                         "start and success messages should be reported")

    @simple_layout('layouts/gerrit-checks-scheme.yaml')
    @skipIfMultiScheduler()
    # This is the only gerrit checks API test which is failing because
    # it uses a check scheme rather than an UUID. The scheme must first
    # be evaluated and mapped to an UUID.
    # This shouldn't be a problem in production as the evaluation takes
    # place on the gerrit webserver. However, in the tests we get a
    # dedicated (fake) gerrit webserver for each fake gerrrit
    # connection. Since each scheduler gets a new connection, only one
    # of those webservers will be aware of the check. If any other
    # webserver tries to evaluate the check it will fail with
    # "Unable to find matching checker".
    def test_check_pipeline_scheme(self):
        self.fake_gerrit.addFakeChecker(uuid='zuul_check:abcd',
                                        repository='org/project',
                                        status='ENABLED')
        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
        self.waitUntilSettled()

        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        A.setCheck('zuul_check:abcd', reset=True)
        self.waitForPoll('gerrit')
        self.waitUntilSettled()

        self.assertEqual(A.checks_history[0]['zuul_check:abcd']['state'],
                         'NOT_STARTED')
        self.assertEqual(A.checks_history[1]['zuul_check:abcd']['state'],
                         'SCHEDULED')
        self.assertEqual(A.checks_history[2]['zuul_check:abcd']['state'],
                         'RUNNING')
        self.assertEqual(A.checks_history[3]['zuul_check:abcd']['state'],
                         'SUCCESSFUL')
        self.assertEqual(len(A.checks_history), 4)
        self.assertHistory([
            dict(name='test-job', result='SUCCESS', changes='1,1')])

    @simple_layout('layouts/gerrit-checks-nojobs.yaml')
    def test_no_jobs(self):
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        A.setCheck('zuul:check', reset=True)
        self.waitForPoll('gerrit')
        self.waitUntilSettled()

        self.assertEqual(A.checks_history[0]['zuul:check']['state'],
                         'NOT_STARTED')
        self.assertEqual(A.checks_history[1]['zuul:check']['state'],
                         'SCHEDULED')
        self.assertEqual(A.checks_history[2]['zuul:check']['state'],
                         'NOT_RELEVANT')
        self.assertEqual(len(A.checks_history), 3)
        self.assertEqual(A.data['status'], 'NEW')

    @simple_layout('layouts/gerrit-checks.yaml')
    def test_config_error(self):
        # Test that line comments are reported on config errors
        in_repo_conf = textwrap.dedent(
            """
            - project:
                check:
                  jobs:
                    - bad-job
            """)
        file_dict = {'.zuul.yaml': in_repo_conf}
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A',
                                           files=file_dict)
        A.setCheck('zuul:check', reset=True)
        self.waitForPoll('gerrit')
        self.waitUntilSettled()

        self.assertEqual(A.checks_history[0]['zuul:check']['state'],
                         'NOT_STARTED')
        self.assertEqual(A.checks_history[1]['zuul:check']['state'],
                         'SCHEDULED')
        self.assertEqual(A.checks_history[2]['zuul:check']['state'],
                         'FAILED')
        self.assertEqual(len(A.checks_history), 3)
        comments = sorted(A.comments, key=lambda x: x['line'])
        self.assertEqual(comments[0],
                         {'file': '.zuul.yaml',
                          'line': 5,
                          'message': 'Job bad-job not defined',
                          'range': {'end_character': 0,
                                    'end_line': 5,
                                    'start_character': 2,
                                    'start_line': 2},
                          'reviewer': {'email': 'zuul@example.com',
                                       'name': 'Zuul',
                                       'username': 'jenkins'}}
        )
        self.assertEqual(A.reported, 0, "no messages should be reported")
        self.assertEqual(A.messages, [], "no messages should be reported")

    @simple_layout('layouts/gerrit-checks.yaml')
    def test_new_patchset(self):
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        A.setCheck('zuul:check', reset=True)
        self.waitForPoll('gerrit')
        self.waitUntilSettled()

        self.assertEqual(A.checks_history[0]['zuul:check']['state'],
                         'NOT_STARTED')
        self.assertEqual(A.checks_history[1]['zuul:check']['state'],
                         'SCHEDULED')
        self.assertEqual(A.checks_history[2]['zuul:check']['state'],
                         'RUNNING')
        self.assertEqual(len(A.checks_history), 3)

        A.addPatchset()
        A.setCheck('zuul:check', reset=True)
        self.waitForPoll('gerrit')
        self.waitUntilSettled()

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()
        self.log.info(A.checks_history)
        self.assertEqual(A.checks_history[3]['zuul:check']['state'],
                         'NOT_STARTED')
        self.assertEqual(A.checks_history[4]['zuul:check']['state'],
                         'SCHEDULED')
        self.assertEqual(A.checks_history[5]['zuul:check']['state'],
                         'RUNNING')
        self.assertEqual(A.checks_history[6]['zuul:check']['state'],
                         'SUCCESSFUL')
        self.assertEqual(len(A.checks_history), 7)
        self.assertHistory([
            dict(name='test-job', result='ABORTED', changes='1,1'),
            dict(name='test-job', result='SUCCESS', changes='1,2'),
        ], ordered=False)


class TestPolling(ZuulTestCase):
    config_file = 'zuul-gerrit-no-stream.conf'

    @simple_layout('layouts/gerrit-checks.yaml')
    def test_config_update(self):
        # Test that the config is updated via polling when a change
        # merges without stream-events enabled.
        in_repo_conf = textwrap.dedent(
            """
            - job:
                name: test-job2
                parent: test-job
            - project:
                check:
                  jobs:
                    - test-job2
            """)

        file_dict = {'.zuul.yaml': in_repo_conf}
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A',
                                           files=file_dict)
        A.setMerged()
        self.waitForPoll('gerrit')
        self.waitUntilSettled('poll 1')

        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        B.setCheck('zuul:check', reset=True)
        self.waitForPoll('gerrit')
        self.waitUntilSettled('poll 2')

        self.assertEqual(B.checks_history[0]['zuul:check']['state'],
                         'NOT_STARTED')
        self.assertEqual(B.checks_history[1]['zuul:check']['state'],
                         'SCHEDULED')
        self.assertEqual(B.checks_history[2]['zuul:check']['state'],
                         'RUNNING')
        self.assertEqual(B.checks_history[3]['zuul:check']['state'],
                         'SUCCESSFUL')
        self.assertEqual(len(B.checks_history), 4)
        self.assertHistory([
            dict(name='test-job', result='SUCCESS', changes='2,1'),
            dict(name='test-job2', result='SUCCESS', changes='2,1'),
        ], ordered=False)

        # There may be an extra change-merged event because we poll
        # too often during the tests.
        self.waitForPoll('gerrit')
        self.waitUntilSettled('poll 3')

    @simple_layout('layouts/gerrit-poll-post.yaml')
    def test_post(self):
        # Test that ref-updated events trigger post jobs.
        self.waitUntilSettled()
        # Wait for an initial poll to get the original sha.
        self.waitForPoll('gerrit-ref')

        # Merge a change.
        self.create_commit('org/project')

        # Wait for the job to run.
        self.waitForPoll('gerrit-ref')
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='post-job', result='SUCCESS'),
        ])

    @simple_layout('layouts/gerrit-poll-post.yaml')
    def test_tag(self):
        # Test that ref-updated events trigger post jobs.
        self.waitUntilSettled()
        # Wait for an initial poll to get the original sha.
        self.waitForPoll('gerrit-ref')

        # Merge a change.
        self.fake_gerrit.addFakeTag('org/project', 'master', 'foo')

        # Wait for the job to run.
        self.waitForPoll('gerrit-ref')
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='tag-job', result='SUCCESS'),
        ])


class TestWrongConnection(ZuulTestCase):
    config_file = 'zuul-connections-multiple-gerrits.conf'
    tenant_config_file = 'config/wrong-connection-in-pipeline/main.yaml'

    def test_wrong_connection(self):
        # Test if the wrong connection is configured in a gate pipeline

        # Our system has two gerrits, and we have configured a gate
        # pipeline to trigger on the "review_gerrit" connection, but
        # report (and merge) via "another_gerrit".
        A = self.fake_review_gerrit.addFakeChange('org/project', 'master', 'A')
        A.addApproval('Code-Review', 2)
        self.fake_review_gerrit.addEvent(A.addApproval('Approved', 1))

        self.waitUntilSettled()

        B = self.fake_review_gerrit.addFakeChange('org/project', 'master', 'B')
        # Let's try this as if the change was merged (say, via another tenant).
        B.setMerged()
        B.addApproval('Code-Review', 2)
        self.fake_review_gerrit.addEvent(B.addApproval('Approved', 1))

        self.waitUntilSettled()

        self.assertEqual(A.reported, 0)
        self.assertEqual(B.reported, 0)
        self.assertHistory([
            dict(name='test-job', result='SUCCESS', changes='1,1'),
            dict(name='test-job', result='SUCCESS', changes='2,1'),
        ], ordered=False)


class TestGerritFake(ZuulTestCase):
    config_file = "zuul-gerrit-github.conf"
    tenant_config_file = "config/circular-dependencies/main.yaml"

    def _make_tuple(self, data):
        ret = []
        for c in data:
            dep_change = c['number']
            dep_ps = c['currentPatchSet']['number']
            ret.append((int(dep_change), int(dep_ps)))
        return sorted(ret)

    def _get_tuple(self, change_number):
        ret = []
        data = self.fake_gerrit.get(
            f'changes/{change_number}/submitted_together')
        for c in data:
            dep_change = c['_number']
            dep_ps = c['revisions'][c['current_revision']]['_number']
            ret.append((dep_change, dep_ps))
        return sorted(ret)

    def test_submitted_together_normal(self):
        # Test that the fake submitted together endpoint returns
        # expected data

        # This test verifies behavior with submitWholeTopic=False

        # A single change
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        data = self._get_tuple(1)
        self.assertEqual(data, [])

        # A dependent series (B->A)
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        B.setDependsOn(A, 1)
        data = self._get_tuple(2)
        self.assertEqual(data, [(1, 1), (2, 1)])

        # A topic cycle
        self.fake_gerrit.addFakeChange('org/project', 'master', 'C1',
                                       topic='test-topic')
        self.fake_gerrit.addFakeChange('org/project', 'master', 'C2',
                                       topic='test-topic')
        data = self._get_tuple(3)
        self.assertEqual(data, [])

    @gerrit_config(submit_whole_topic=True)
    def test_submitted_together_whole_topic(self):
        # Test that the fake submitted together endpoint returns
        # expected data
        # This test verifies behavior with submitWholeTopic=True

        # A single change
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        data = self._get_tuple(1)
        self.assertEqual(data, [])

        # A dependent series (B->A)
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        B.setDependsOn(A, 1)
        data = self._get_tuple(2)
        self.assertEqual(data, [(1, 1), (2, 1)])
        # The Gerrit connection method filters out the queried change

        # A topic cycle
        self.fake_gerrit.addFakeChange('org/project', 'master', 'C1',
                                       topic='test-topic')
        self.fake_gerrit.addFakeChange('org/project', 'master', 'C2',
                                       topic='test-topic')
        data = self._get_tuple(3)
        self.assertEqual(data, [(3, 1), (4, 1)])
        # The Gerrit connection method filters out the queried change

        # Test also the query used by the GerritConnection:
        ret = self.fake_gerrit._simpleQuery('status:open topic:test-topic')
        ret = self._make_tuple(ret)
        self.assertEqual(ret, [(3, 1), (4, 1)])


class TestGerritConnection(ZuulTestCase):
    config_file = 'zuul-gerrit-web.conf'
    tenant_config_file = 'config/single-tenant/main.yaml'

    def test_zuul_query_ltime(self):
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        B.setDependsOn(A, 1)

        # We want the three events below to proceed so that all three
        # events are processed simultaneously by three threads.  We
        # will allow the first event we submit to complete its change
        # network and therefore perform all the queries.  The other
        # two events should receive ChangeNetworkConflict errors and
        # eventually use the results of the first network.  To
        # accomplish that, we: Allow the network to proceed if it
        # includes change A (this will be our first network) * Wait
        # for two ChangeNetworkConflict errors to occur
        permission_barrier = threading.Barrier(2)
        permission_event = threading.Event()
        cn_manager = self.fake_gerrit.change_network_manager
        orig_permission = cn_manager.permissionToProceed

        def permissionToProceed(future):
            if permission_event.is_set():
                # We're past the critical section, revert to normal
                # behavior.
                return orig_permission(future)
            try:
                changes = [int(c.stable_id) for c in future.changes]
                if 1 in changes:
                    # This change network includes change A; this is a
                    # characteristic of the first event, so we let it
                    # through.
                    return orig_permission(future)
                for _ in iterate_timeout(30, "queries to occur"):
                    # Wait until the first change network has expanded
                    # to include both changes so that it will
                    # conflict.
                    for f in cn_manager.futures:
                        if len(f.changes) > 1:
                            return orig_permission(future)
            except ChangeNetworkConflict:
                # The second and third events will conflict; once the
                # barrier tells us that they have both gotten to this
                # point, we can allow the first network to finish and
                # these two threads to resume normal operation (they
                # will use the results of their failed networks to try
                # again.
                permission_barrier.wait()
                permission_event.set()
                raise

        self.patch(self.fake_gerrit.change_network_manager,
                   'permissionToProceed', permissionToProceed)

        orig_complete = cn_manager.setComplete

        def setComplete(future):
            # The first change network must wait here until the second
            # and third have seen conflict errors from it, so that it
            # is not removed from the manager before they see the
            # conflict.
            permission_event.wait()
            return orig_complete(future)

        self.patch(self.fake_gerrit.change_network_manager,
                   'setComplete', setComplete)

        # These are the three events with critical ordering described
        # above.
        self.fake_gerrit.addEvent(A.addApproval('Code-Review', 2))
        self.fake_gerrit.addEvent(B.addApproval('Approved', 1))
        self.fake_gerrit.addEvent(B.addApproval('Code-Review', 2))

        self.waitUntilSettled()
        self.assertHistory([])
        # One query for each change in the above cluster of events.
        self.assertEqual(A.queried, 1)
        self.assertEqual(B.queried, 1)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()
        self.assertHistory([
            dict(name="project-merge", result="SUCCESS", changes="1,1"),
            dict(name="project-test1", result="SUCCESS", changes="1,1"),
            dict(name="project-test2", result="SUCCESS", changes="1,1"),
            dict(name="project-merge", result="SUCCESS", changes="1,1 2,1"),
            dict(name="project-test1", result="SUCCESS", changes="1,1 2,1"),
            dict(name="project-test2", result="SUCCESS", changes="1,1 2,1"),
        ], ordered=False)
        # One query due to the event on change A, followed by a query
        # to verify the merge.
        self.assertEqual(A.queried, 3)
        # No query for change B necessary since our cache is up to
        # date with respect for the triggering event.  One query to
        # verify the merge.
        self.assertEqual(B.queried, 2)
        self.assertEqual(A.data['status'], 'MERGED')
        self.assertEqual(B.data['status'], 'MERGED')

    def test_submit_requirements(self):
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        A.addApproval('Code-Review', 2)
        # Set an unsatisfied submit requirement
        A.setSubmitRequirements([
            {
                "name": "Code-Review",
                "description": "Disallow self-review",
                "status": "UNSATISFIED",
                "is_legacy": False,
                "submittability_expression_result": {
                    "expression": "label:Code-Review=MAX,user=non_uploader "
                        "AND -label:Code-Review=MIN",
                    "fulfilled": False,
                    "passing_atoms": [],
                    "failing_atoms": [
                        "label:Code-Review=MAX,user=non_uploader",
                        "label:Code-Review=MIN"
                    ]
                }
            },
            {
                "name": "Verified",
                "status": "UNSATISFIED",
                "is_legacy": True,
                "submittability_expression_result": {
                    "expression": "label:Verified=MAX -label:Verified=MIN",
                    "fulfilled": False,
                    "passing_atoms": [],
                    "failing_atoms": [
                        "label:Verified=MAX",
                        "-label:Verified=MIN"
                    ]
                }
            },
        ])
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()
        self.assertHistory([])
        self.assertEqual(A.queried, 1)
        self.assertEqual(A.data['status'], 'NEW')

        # Mark the requirement satisfied
        A.setSubmitRequirements([
            {
                "name": "Code-Review",
                "description": "Disallow self-review",
                "status": "SATISFIED",
                "is_legacy": False,
                "submittability_expression_result": {
                    "expression": "label:Code-Review=MAX,user=non_uploader "
                        "AND -label:Code-Review=MIN",
                    "fulfilled": False,
                    "passing_atoms": [
                        "label:Code-Review=MAX,user=non_uploader",
                    ],
                    "failing_atoms": [
                        "label:Code-Review=MIN"
                    ]
                }
            },
            {
                "name": "Verified",
                "status": "UNSATISFIED",
                "is_legacy": True,
                "submittability_expression_result": {
                    "expression": "label:Verified=MAX -label:Verified=MIN",
                    "fulfilled": False,
                    "passing_atoms": [],
                    "failing_atoms": [
                        "label:Verified=MAX",
                        "-label:Verified=MIN"
                    ]
                }
            },
        ])
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()
        self.assertHistory([
            dict(name="project-merge", result="SUCCESS", changes="1,1"),
            dict(name="project-test1", result="SUCCESS", changes="1,1"),
            dict(name="project-test2", result="SUCCESS", changes="1,1"),
        ], ordered=False)
        self.assertEqual(A.queried, 3)
        self.assertEqual(A.data['status'], 'MERGED')

    def test_submit_missing_labels(self):
        # This test checks that missing labels will not allow a change to
        # enqueued. Specifically in this case the Code-Review and Verified is
        # missing (but Verified is ignored because Zuul will score it)
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()
        self.assertHistory([])

        # Even adding the Verified score already provided by Zuul will not
        # cause the change to gate
        self.fake_gerrit.addEvent(A.addApproval('Verified', 2))

        # Reapply this score to trigger gate, which shouldn't happen because of
        # the still missing Code-Review label
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()
        self.assertHistory([])

    def test_query_abandoned_changes_check(self):
        # Test that we stop querying when we encounter an abandoned change
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        C = self.fake_gerrit.addFakeChange('org/project', 'master', 'C')
        D = self.fake_gerrit.addFakeChange('org/project', 'master', 'D')

        # D(open) -> C(abandoned) -> B(abandoned) -> A(open)
        D.setDependsOn(C, 1)
        C.setDependsOn(B, 1)
        B.setDependsOn(A, 1)
        C.setAbandoned()
        B.setAbandoned()
        self.fake_gerrit.addEvent(D.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.assertEqual(A.queried, 0)
        self.assertEqual(B.queried, 0)
        self.assertEqual(C.queried, 1)
        self.assertEqual(D.queried, 1)

        self.assertHistory([
            dict(name="project-merge", result="SUCCESS", changes="3,1 4,1"),
            dict(name="project-test1", result="SUCCESS", changes="3,1 4,1"),
            dict(name="project-test2", result="SUCCESS", changes="3,1 4,1"),
        ], ordered=False)

    def test_query_abandoned_changes_gate(self):
        # Test that we stop querying when we encounter an abandoned change
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        C = self.fake_gerrit.addFakeChange('org/project', 'master', 'C')
        D = self.fake_gerrit.addFakeChange('org/project', 'master', 'D')

        # D(open) -> C(abandoned) -> B(abandoned) -> A(open)
        D.setDependsOn(C, 1)
        C.setDependsOn(B, 1)
        B.setDependsOn(A, 1)
        C.setAbandoned()
        B.setAbandoned()
        A.addApproval('Code-Review', 2)
        A.addApproval('Approved', 1)
        B.addApproval('Code-Review', 2)
        B.addApproval('Approved', 1)
        C.addApproval('Code-Review', 2)
        C.addApproval('Approved', 1)
        D.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(D.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertEqual(A.queried, 0)
        self.assertEqual(B.queried, 0)
        self.assertEqual(C.queried, 1)
        self.assertEqual(D.queried, 1)

        self.assertHistory([])


class TestGerritConnectionDelay(ZuulTestCase):
    config_file = 'zuul-gerrit-web-delay.conf'
    tenant_config_file = 'config/single-tenant/main.yaml'
    replication_delay = 10

    def test_ref_updated_reconfig(self):
        # Gerrit emits change-merged events after ref-updated events for the
        # change; make sure that job configuration changes take effect
        # for post pipelines that trigger off of ref-updated.
        in_repo_conf = textwrap.dedent(
            """
            - job:
                name: new-post-job
                parent: project-post
            - project:
                post:
                  jobs:
                    - new-post-job
            """)

        file_dict = {'zuul.yaml': in_repo_conf}
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A',
                                           files=file_dict)
        A.setMerged()
        self.fake_gerrit.addEvent(A.getRefUpdatedEvent())
        # We add a delay here so that we exercise the behavior at the end
        # of the ZK queue.
        time.sleep(1)
        self.fake_gerrit.addEvent(A.getChangeMergedEvent())
        self.waitUntilSettled()
        self.assertHistory([
            dict(name='project-post', result='SUCCESS'),
            dict(name='new-post-job', result='SUCCESS'),
        ], ordered=False)

    def test_ref_updated_reconfig_immediate(self):
        # Same as the above test, but with no delay between the two
        # events.
        in_repo_conf = textwrap.dedent(
            """
            - job:
                name: new-post-job
                parent: project-post
            - project:
                post:
                  jobs:
                    - new-post-job
            """)

        file_dict = {'zuul.yaml': in_repo_conf}
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A',
                                           files=file_dict)
        A.setMerged()
        self.fake_gerrit.addEvent(A.getRefUpdatedEvent())
        self.fake_gerrit.addEvent(A.getChangeMergedEvent())
        self.waitUntilSettled()
        self.assertHistory([
            dict(name='project-post', result='SUCCESS'),
            dict(name='new-post-job', result='SUCCESS'),
        ], ordered=False)

    def test_ref_updated_timeout(self):
        # Test the behavior when the second event is missed
        self.hold_merge_jobs_in_queue = True
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        A.setMerged()
        start = time.time()
        self.fake_gerrit.addEvent(A.getRefUpdatedEvent())
        self.waitUntilSettled()
        end = time.time()
        elapsed = end - start
        self.log.debug("Elapsed time: %s", elapsed)
        # We are waiting replication_delay + peekqueue timeout to pass
        # for the corresponding change merged event that matches our
        # ref updated event.
        self.assertTrue(10 + self.replication_delay
                        <= elapsed <
                        20 + self.replication_delay)
        self.hold_merge_jobs_in_queue = False
        self.merger_api.release()
        self.waitUntilSettled()

    def test_replication_delay_one_change(self):
        # Test that a single change uses the correct replication delay
        self.hold_merge_jobs_in_queue = True
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        start = time.time()
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        end = time.time()
        elapsed = end - start
        self.log.debug("Elapsed time: %s", elapsed)
        self.assertTrue(self.replication_delay
                        <= elapsed <
                        self.replication_delay + 10)
        self.hold_merge_jobs_in_queue = False
        self.merger_api.release()
        self.waitUntilSettled()

    def test_replication_delay_two_changes(self):
        # Test that two changes are collectively delayed only 10 seconds
        self.hold_merge_jobs_in_queue = True
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        start = time.time()
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        end = time.time()
        elapsed = end - start
        self.log.debug("Elapsed time: %s", elapsed)
        self.assertTrue(self.replication_delay
                        <= elapsed <
                        self.replication_delay + 10)
        self.hold_merge_jobs_in_queue = False
        self.merger_api.release()
        self.waitUntilSettled()


class TestGerritConnectionNoDelay(TestGerritConnectionDelay):
    config_file = 'zuul-gerrit-web.conf'
    tenant_config_file = 'config/single-tenant/main.yaml'
    replication_delay = 0


class TestGerritConnectionPreFilter(ZuulTestCase):
    config_file = 'zuul-gerrit-web.conf'
    tenant_config_file = 'config/pre-filter/main.yaml'

    def test_ignored_events(self):
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        event_queue = self.fake_gerrit.gerrit_event_connector.event_queue

        # Reconfigure the second tenant to ensure that it doesn't
        # delete the pre-filters for the first tenant.
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-two')
        (trusted, project1) = tenant.getProject('org/project2')
        event = zuul.model.TriggerEvent()
        event.zuul_event_ltime = -1
        self.scheds.first.sched.reconfigureTenant(
            self.scheds.first.sched.abide.tenants['tenant-two'],
            project1, event)
        self.waitUntilSettled()

        # For efficiency, ensure that the total filter length is the
        # number of filters in the first tenant plus 1 (there is one
        # unique filter in the second tenant).
        self.assertEqual(6, len(self.fake_gerrit.watched_event_filters))

        # Stop the event connector so it does not pull any events from
        # the queue
        self.fake_gerrit.stopEventConnector()
        # This is ignored unconditionally
        self.fake_gerrit.addEvent({
            "type": "cache-eviction",
        })
        self.assertEqual(0, len(event_queue._listEvents()))
        # This is ignored for meta refs only
        self.fake_gerrit.addEvent({
            "type": "ref-updated",
            "submitter": {
                "name": "User Name",
            },
            "refUpdate": {
                "oldRev": '0',
                "newRev": '0',
                "refName": 'refs/changes/01/1/meta',
                "project": 'org/project1',
            }
        })
        self.assertEqual(0, len(event_queue._listEvents()))
        # This is ignored because the ref doesn't match the pre-filter
        self.fake_gerrit.addEvent({
            "type": "ref-updated",
            "submitter": {
                "name": "User Name",
            },
            "refUpdate": {
                "oldRev": '0',
                "newRev": '0',
                "refName": 'refs/something/else',
                "project": 'org/project1',
            }
        })
        self.assertEqual(0, len(event_queue._listEvents()))
        # This is not ignored
        self.fake_gerrit.addEvent({
            "type": "ref-updated",
            "submitter": {
                "name": "User Name",
            },
            "refUpdate": {
                "oldRev": '0',
                "newRev": 'abcd',
                "refName": 'refs/tags/1234',
                "project": 'org/project1',
            }
        })
        self.assertEqual(1, len(event_queue._listEvents()))
        # This is not ignored
        A.setMerged()
        self.fake_gerrit.addEvent(A.getRefUpdatedEvent())
        self.assertEqual(2, len(event_queue._listEvents()))

        self.fake_gerrit.startEventConnector()
        self.waitUntilSettled()


class TestGerritUnicodeRefs(ZuulTestCase):
    config_file = 'zuul-gerrit-web.conf'
    tenant_config_file = 'config/single-tenant/main.yaml'

    upload_pack_data = (b'014452944ee370db5c87691e62e0f9079b6281319b4e HEAD'
                        b'\x00multi_ack thin-pack side-band side-band-64k '
                        b'ofs-delta shallow deepen-since deepen-not '
                        b'deepen-relative no-progress include-tag '
                        b'multi_ack_detailed allow-tip-sha1-in-want '
                        b'allow-reachable-sha1-in-want '
                        b'symref=HEAD:refs/heads/faster filter '
                        b'object-format=sha1 agent=git/2.37.1.gl1\n'
                        b'003d5f42665d737b3fd4ec22ca0209e6191859f09fd6 '
                        b'refs/for/faster\n'
                        b'004952944ee370db5c87691e62e0f9079b6281319b4e '
                        b'refs/heads/foo/\xf0\x9f\x94\xa5\xf0\x9f\x94\xa5'
                        b'\xf0\x9f\x94\xa5\n'
                        b'003f52944ee370db5c87691e62e0f9079b6281319b4e '
                        b'refs/heads/faster\n0000').decode("utf-8")

    def test_mb_unicode_refs(self):
        gerrit_config = {
            'user': 'gerrit',
            'server': 'localhost',
        }
        driver = GerritDriver()
        gerrit = GerritConnection(driver, 'review_gerrit', gerrit_config)

        def _uploadPack(project):
            return self.upload_pack_data

        self.patch(gerrit, '_uploadPack', _uploadPack)

        project = gerrit.source.getProject('org/project')
        refs = gerrit.getInfoRefs(project)

        self.assertEqual(refs,
                         {'refs/for/faster':
                          '5f42665d737b3fd4ec22ca0209e6191859f09fd6',
                          'refs/heads/foo/🔥🔥🔥':
                          '52944ee370db5c87691e62e0f9079b6281319b4e',
                          'refs/heads/faster':
                          '52944ee370db5c87691e62e0f9079b6281319b4e'})


class TestGerritDefaultBranch(ZuulTestCase):
    config_file = 'zuul-gerrit-web.conf'
    tenant_config_file = 'config/single-tenant/main.yaml'
    scheduler_count = 1

    def test_default_branch_changed(self):
        """Test the project-head-updated event"""
        self.waitUntilSettled()

        layout = self.scheds.first.sched.abide.tenants.get('tenant-one').layout
        md = layout.getProjectMetadata('review.example.com/org/project1')
        self.assertEqual('master', md.default_branch)
        prev_layout = layout.uuid

        self.fake_gerrit._fake_project_default_branch[
            'org/project1'] = 'foobar'
        self.fake_gerrit.addEvent(self.fake_gerrit.getProjectHeadUpdatedEvent(
            project='org/project1',
            old='master',
            new='foobar'
        ))
        self.waitUntilSettled()

        layout = self.scheds.first.sched.abide.tenants.get('tenant-one').layout
        md = layout.getProjectMetadata('review.example.com/org/project1')
        self.assertEqual('foobar', md.default_branch)
        new_layout = layout.uuid
        self.assertNotEqual(new_layout, prev_layout)


class TestGerritMaxDeps(ZuulTestCase):
    config_file = 'zuul-gerrit-max-deps.conf'

    @simple_layout('layouts/simple.yaml')
    def test_max_deps(self):
        # max_dependencies for the connection is 1, so this is okay
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='check-job', result='SUCCESS', changes='1,1'),
        ])

        # So is this
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        B.setDependsOn(A, 1)
        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='check-job', result='SUCCESS', changes='1,1'),
            dict(name='check-job', result='SUCCESS', changes='1,1 2,1'),
        ])

        # This is not
        C = self.fake_gerrit.addFakeChange('org/project', 'master', 'C')
        C.data['commitMessage'] = '%s\n\nDepends-On: %s\nDepends-On: %s\n' % (
            C.subject, A.data['id'], B.data['id'])
        self.fake_gerrit.addEvent(C.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='check-job', result='SUCCESS', changes='1,1'),
            dict(name='check-job', result='SUCCESS', changes='1,1 2,1'),
        ])


class TestGerritDriver(ZuulTestCase):
    # Most of the Zuul test suite tests the Gerrit driver, to some
    # extent.  The other classes in this file test specific methods of
    # Zuul interacting with Gerrit.  But the other drivers also test
    # some basic driver functionality that, if tested for Gerrit at
    # all, is spread out in random tests.  This class adds some
    # (potentially duplicative) testing to validate parity with the
    # other drivers.

    @simple_layout('layouts/simple.yaml')
    def test_change_event(self):
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A',
                                           topic='test-topic')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='check-job', result='SUCCESS', changes='1,1'),
        ])

        job = self.getJobFromHistory('check-job')
        zuulvars = job.parameters['zuul']
        self.assertEqual(str(A.number), zuulvars['change'])
        self.assertEqual('1', zuulvars['patchset'])
        self.assertEqual(str(A.patchsets[-1]['revision']),
                         zuulvars['commit_id'])
        self.assertEqual('master', zuulvars['branch'])
        self.assertEqual('https://review.example.com/1',
                         zuulvars['items'][0]['change_url'])
        self.assertEqual(zuulvars["message"], strings.b64encode('A'))
        self.assertEqual(1, len(self.history))
        self.assertEqual(1, len(A.messages))
        self.assertEqual('test-topic', zuulvars['topic'])

    @simple_layout('layouts/simple.yaml')
    def test_tag_event(self):
        event = self.fake_gerrit.addFakeTag('org/project', 'master', 'foo')
        tagsha = event['refUpdate']['newRev']
        self.fake_gerrit.addEvent(event)
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='tag-job', result='SUCCESS', ref='refs/tags/foo'),
        ])

        job = self.getJobFromHistory('tag-job')
        zuulvars = job.parameters['zuul']
        zuulvars = job.parameters['zuul']
        self.assertEqual('refs/tags/foo', zuulvars['ref'])
        self.assertEqual('tag', zuulvars['pipeline'])
        self.assertEqual('tag-job', zuulvars['job'])
        self.assertEqual(tagsha, zuulvars['newrev'])
        self.assertEqual(tagsha, zuulvars['commit_id'])

    @simple_layout('layouts/gerrit-hashtags.yaml')
    def test_hashtags_event(self):
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        A.data['hashtags'] = ['check']
        self.fake_gerrit.addEvent(A.getHashtagsChangedEvent(added=['check']))
        self.waitUntilSettled()

        # Does not meet pipeline requirement
        self.assertHistory([])

        A.data['hashtags'] = ['okay', 'check']
        self.fake_gerrit.addEvent(A.getHashtagsChangedEvent(added=['check']))
        self.waitUntilSettled()

        # This should work
        self.assertHistory([
            dict(name='check-job', result='SUCCESS', changes='1,1'),
        ])

        # Matches reject
        A.data['hashtags'] = ['okay', 'check', 'nope']
        self.fake_gerrit.addEvent(A.getHashtagsChangedEvent(
            removed=['nocheck']))
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='check-job', result='SUCCESS', changes='1,1'),
        ])

    @simple_layout('layouts/gerrit-approval-change.yaml')
    def test_approval_change(self):
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        A.data['hashtags'] = ['check']
        self.fake_gerrit.addEvent(A.addApproval('Code-Review', 2))
        self.waitUntilSettled()

        # Does not meet pipeline requirements, because old value is not present
        self.assertHistory([])

        # Still not sufficient because old value matches new value
        self.fake_gerrit.addEvent(A.addApproval('Code-Review', 2, old_value=2))
        self.waitUntilSettled()
        self.assertHistory([])

        # Old value present, but new value is insufficient
        self.fake_gerrit.addEvent(A.addApproval('Code-Review', 0, old_value=2))
        self.waitUntilSettled()
        self.assertHistory([])

        # This should work now
        self.fake_gerrit.addEvent(A.addApproval('Code-Review', 2, old_value=0))
        self.waitUntilSettled()
        self.assertHistory([
            dict(name='check-job', result='SUCCESS', changes='1,1'),
        ])

    @simple_layout('layouts/gerrit-notify.yaml')
    def test_notify(self):
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='check-job', result='SUCCESS'),
        ])

        self.assertEqual(A.notify, 'NONE')


class TestGerritCherryPickSSH(ZuulTestCase):
    @simple_layout('layouts/cherry-pick.yaml')
    def test_gerrit_cherry_pick_ssh(self):
        # Test the simple case case where we approve A then B.
        # Because the gerrit driver doesn't store the new patchset
        # that is created on a cherry-pick at the time of submission,
        # this test actually works because we intentionally
        # erroneously mark the old revision as merged.  So when we
        # enqueue B, it sees its dependency of 1,1 as merged (even
        # though 1,2 was what was actually created-and-merged.  This
        # test is here to ensure that works.
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        A.cherry_pick = True
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        B.cherry_pick = True
        B.setDependsOn(A, 1)
        A.addApproval('Code-Review', 2)
        B.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='check-job', result='SUCCESS', changes='1,1'),
        ])

        self.fake_gerrit.addEvent(B.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='check-job', result='SUCCESS', changes='1,1'),
            dict(name='check-job', result='SUCCESS', changes='2,1'),
        ])
    # Unlike the web test below, we are unable to handle a case where
    # we add a patchset to change A before enqueing B, because the ssh
    # query does not tell us the current patchset of dependencies, so
    # there is no corresponding test for the ssh code path.


class TestGerritCherryPickWeb(ZuulTestCase):
    config_file = 'zuul-gerrit-web.conf'

    @simple_layout('layouts/cherry-pick.yaml')
    def test_gerrit_cherry_pick_web(self):
        # Test the simple case case where we approve A then B.
        # Because the gerrit driver doesn't store the new patchset
        # that is created on a cherry-pick at the time of submission,
        # this test actually works because we intentionally
        # erroneously mark the old revision as merged.  So when we
        # enqueue B, it sees its dependency of 1,1 as merged (even
        # though 1,2 was what was actually created-and-merged.  This
        # test is here to ensure that works.
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        A.cherry_pick = True
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        B.cherry_pick = True
        B.setDependsOn(A, 1)
        A.addApproval('Code-Review', 2)
        B.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='check-job', result='SUCCESS', changes='1,1'),
        ])

        self.fake_gerrit.addEvent(B.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='check-job', result='SUCCESS', changes='1,1'),
            dict(name='check-job', result='SUCCESS', changes='2,1'),
        ])

    @simple_layout('layouts/cherry-pick.yaml')
    def test_gerrit_cherry_pick_web_with_update(self):
        # Test updating A behind B's back then approving B.
        # Unlike the test above, this test forces the issue of
        # patchset updates by having the user create a second patchset
        # on A before anything is enqueued.  But B still depends on
        # the first patchset of A.  So when the gerrit driver
        # intentionally erroneously marks 1,2 as merged, that won't
        # help B since it points at 1,1.  However, the web query code
        # path will update B's dependency to point at 1,2 in the case
        # of a cherry-pick, and we will query 1,2 and discover that it
        # is merged, so B can proceed.
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        A.cherry_pick = True
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        B.cherry_pick = True
        # Set the dependency on the first patchset of A
        B.setDependsOn(A, 1)
        # Then add a new patchset to A, so that B points to an old
        # version.
        A.addPatchset()
        A.addApproval('Code-Review', 2)
        B.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='check-job', result='SUCCESS', changes='1,2'),
        ])

        self.fake_gerrit.addEvent(B.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='check-job', result='SUCCESS', changes='1,2'),
            dict(name='check-job', result='SUCCESS', changes='2,1'),
        ])

    @simple_layout('layouts/cherry-pick.yaml')
    def test_gerrit_cherry_pick_web_with_update_queue(self):
        # Test that two cherry-pick changes enqued simultaneously work
        # This test should exercise the code path where the second
        # change remains blissfully unaware that there ever was a
        # 1,3 patchset.
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        A.cherry_pick = True
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        B.cherry_pick = True
        # Set the dependency on the first patchset of A
        B.setDependsOn(A, 1)
        # Then add a new patchset to A, so that B points to an old
        # version.
        A.addPatchset()
        A.addApproval('Code-Review', 2)
        B.addApproval('Code-Review', 2)
        B.addApproval('Approved', 1)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='check-job', result='SUCCESS', changes='1,2'),
            dict(name='check-job', result='SUCCESS', changes='1,2 2,1'),
        ], ordered=False)


class TestGerritPeekQueue(BaseTestCase):
    def make_ref_updated_event(self, ref, newrev, ltime):
        e = zuul.model.ConnectionEvent({
            "timestamp": time.time(),
            "payload": {
                "refUpdate": {
                    "oldRev": "old",
                    "newRev": newrev,
                    "refName": ref,
                    "project": "org/project"
                },
                "type": "ref-updated",
            }
        })
        e.zuul_event_ltime = ltime
        return e

    def make_change_merged_event(self, number, newrev, ltime):
        e = zuul.model.ConnectionEvent({
            "timestamp": time.time(),
            "payload": {
                "newRev": newrev,
                "patchSet": {
                    "number": 1,
                },
                "change": {
                    "project": "org/project",
                    "branch": "master",
                    "number": number,
                },
                "type": "change-merged",
            }
        })
        e.zuul_event_ltime = ltime
        return e

    def test_peek_queue(self):
        handled = []

        def handler(x):
            handled.append(x)

        q = PeekQueue(handler, 0)

        # Check noop
        q.run()
        self.assertEqual([], handled)

        # Check all at once (even though this shouldn't happen
        orig = [
            self.make_ref_updated_event('refs/heads/master', 'new1', 1),
            self.make_ref_updated_event('refs/meta/foo', 'foo', 2),
            self.make_change_merged_event(1, 'new1', 3),
        ]
        [q.append(x) for x in orig]
        q.run()
        expected = [orig[2], orig[0], orig[1]]
        self.assertEqual(expected, handled)
        self.assertEqual(1, handled[0].zuul_event_ltime)
        self.assertEqual(1, handled[1].zuul_event_ltime)
        self.assertEqual(2, handled[2].zuul_event_ltime)

        # Check one at a time (typical case)
        handled.clear()
        orig = [
            self.make_ref_updated_event('refs/heads/master', 'new1', 1),
            self.make_ref_updated_event('refs/meta/foo', 'foo', 2),
            self.make_change_merged_event(1, 'new1', 3),
        ]
        q.append(orig[0])
        q.run()
        self.assertEqual([], handled)

        q.append(orig[1])
        q.run()
        self.assertEqual([], handled)

        q.append(orig[2])
        q.run()
        expected = [orig[2], orig[0], orig[1]]
        self.assertEqual(expected, handled)
        self.assertEqual(1, handled[0].zuul_event_ltime)
        self.assertEqual(1, handled[1].zuul_event_ltime)
        self.assertEqual(2, handled[2].zuul_event_ltime)

        # Check that other events arrive immediately
        handled.clear()

        q.append(orig[1])
        q.run()
        expected = [orig[1]]
        self.assertEqual(expected, handled)

        # Check missing change-merged event
        handled.clear()
        orig = [
            self.make_ref_updated_event('refs/heads/master', 'new1', 1),
            self.make_ref_updated_event('refs/meta/foo', 'foo', 2),
        ]
        q.append(orig[0])
        q.run()
        self.assertEqual([], handled)

        q.append(orig[1])
        q.run()
        self.assertEqual([], handled)

        # This is what the loop will acutally do
        q.run()
        # Normally we pass the current time in, but in the test we
        # pass in a future time to accelerate the timout (the peek
        # queue never consults the current time on its own).
        q.run(cutoff=time.time() + 10)
        expected = [orig[0], orig[1]]
        self.assertEqual(expected, handled)
        self.assertEqual(1, handled[0].zuul_event_ltime)
        self.assertEqual(2, handled[1].zuul_event_ltime)

        # Check if Gerrit changes the event order; it doesn't do this
        # today, but we want to defend against that.
        handled.clear()
        orig = [
            self.make_change_merged_event(1, 'new1', 1),
            self.make_ref_updated_event('refs/heads/master', 'new1', 2),
            self.make_ref_updated_event('refs/meta/foo', 'foo', 3),
        ]
        q.append(orig[0])
        q.run()
        expected = [orig[0]]
        self.assertEqual(expected, handled)

        q.append(orig[1])
        q.run()
        expected = [orig[0], orig[1]]
        self.assertEqual(expected, handled)

        q.append(orig[2])
        q.run()
        expected = [orig[0], orig[1], orig[2]]
        self.assertEqual(expected, handled)
        self.assertEqual(1, handled[0].zuul_event_ltime)
        self.assertEqual(2, handled[1].zuul_event_ltime)
        self.assertEqual(3, handled[2].zuul_event_ltime)

    def test_peek_queue_busy_timeout(self):
        # Our last line of defense on a busy system is the extra 10
        # second timeout -- if we see an event that's 10 seconds later
        # than the event we're waiting for, then we'll give up.  If we
        # exhaust the event queue in zookeeper though, we'll give up
        # earlier; that's tested in the normal test_peek_queue test.
        handled = []

        def handler(x):
            handled.append(x)

        q = PeekQueue(handler, 0)

        orig = [
            self.make_ref_updated_event('refs/heads/master', 'new1', 1),
        ]
        [q.append(x) for x in orig]
        q.run()
        expected = []
        self.assertEqual(expected, handled)

        q.timeout = 0
        q.run()
        expected = [orig[0]]
        self.assertEqual(expected, handled)



# Copyright 2014 Rackspace Australia
# Copyright 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import re
import time
from urllib.parse import quote_plus

import alembic
import alembic.command
import alembic.config
import alembic.migration
import sqlalchemy as sa
from sqlalchemy import orm
import sqlalchemy.pool

from zuul.connection import BaseConnection
from zuul.zk.locks import CONNECTION_LOCK_ROOT, locked, SessionAwareLock


BUILDSET_TABLE = 'zuul_buildset'
REF_TABLE = 'zuul_ref'
BUILDSET_REF_TABLE = 'zuul_buildset_ref'
BUILDSET_EVENT_TABLE = 'zuul_buildset_event'
BUILD_TABLE = 'zuul_build'
BUILD_EVENT_TABLE = 'zuul_build_event'
ARTIFACT_TABLE = 'zuul_artifact'
PROVIDES_TABLE = 'zuul_provides'
SYSTEM_EVENT_TABLE = 'zuul_system_event'

POSTGRES_STATEMENT_TIMEOUT_RE = re.compile(
    r'/\* postgres_statement_timeout=(\d+) \*/')
MARIADB_STATEMENT_TIMEOUT_RE = re.compile(
    r'/\* mariadb_statement_timeout=(\d+) \*/')

SQL_MAX_STRING_LENGTH = 255


# In Postgres we can set a per-transaction (which for us is
# effectively per-query) execution timeout by executing "SET LOCAL
# statement_timeout" before our query.  There isn't a great way to do
# this using the SQLAlchemy query API, so instead, we add a comment as
# a hint, and here we parse that comment and execute the "SET".  The
# comment remains in the query sent to the server, but that's okay --
# it may even help an operator in debugging.  The same is true for
# mariadb, except that we need to actually alter the statement instead
# of executing another one.
@sa.event.listens_for(sa.Engine, "before_cursor_execute", retval=True)
def _set_timeout(conn, cursor, stmt, params, context, executemany):
    match = POSTGRES_STATEMENT_TIMEOUT_RE.search(stmt)
    if match:
        cursor.execute("SET LOCAL statement_timeout=%s" % match.groups())
    else:
        match = MARIADB_STATEMENT_TIMEOUT_RE.search(stmt)
        if match:
            stmt = ("SET STATEMENT "
                    f"max_statement_time={float(match.groups()[0]) / 1000} "
                    f"for {stmt}")
    return (stmt, params)


class ChangeType(sa.TypeDecorator):
    """Coerces NULL/None values to/from the integer 0 to facilitate use in
    non-nullable columns

    """
    # Underlying implementation
    impl = sa.Integer
    # We produce consistent values, so can be cached
    cache_ok = True
    # Don't use "IS NULL" when comparing to None values
    coerce_to_is_types = ()

    def process_bind_param(self, value, dialect):
        if value is None:
            return 0
        return value

    def process_result_value(self, value, dialect):
        if value == 0:
            return None
        return value


class SHAType(sa.TypeDecorator):
    """Coerces NULL/None values to/from the empty string to facilitate use
    in non-nullable columns

    """
    # Underlying implementation
    impl = sa.String
    # We produce consistent values, so can be cached
    cache_ok = True
    # Don't use "IS NULL" when comparing to None values
    coerce_to_is_types = ()

    def process_bind_param(self, value, dialect):
        if value is None:
            return ''
        return value

    def process_result_value(self, value, dialect):
        if value == '':
            return None
        return value


class DatabaseSession(object):
    log = logging.getLogger("zuul.DatabaseSession")

    def __init__(self, connection):
        self.connection = connection
        self.session = connection.session

    def __enter__(self):
        return self

    def __exit__(self, etype, value, tb):
        if etype:
            self.session().rollback()
        else:
            self.session().commit()
        self.session().close()
        self.session = None

    def _sanitizeSubstringQuery(self, value):
        if isinstance(value, str):
            escape_char = "$"
            for c in (escape_char, "%", "_"):
                value = value.replace(c, escape_char + c)
            return value
        return value

    def _getFuzzyFilterOp(self, column, value):
        if isinstance(value, str) and "*" in value:
            value = self._sanitizeSubstringQuery(value)
            return column.like(value.replace("*", "%"), escape="$")
        else:
            return column == value

    def listFilterFuzzy(self, query, column, value):
        if value is None:
            return query
        elif isinstance(value, list) or isinstance(value, tuple):
            return query.filter(
                sa.or_(*[self._getFuzzyFilterOp(column, v) for v in value])
            )
        return query.filter(self._getFuzzyFilterOp(column, value))

    def listFilter(self, query, column, value):
        if value is None:
            return query
        if isinstance(value, list) or isinstance(value, tuple):
            return query.filter(column.in_(value))
        return query.filter(column == value)

    def exListFilter(self, query, column, value):
        # Exclude values in list
        if value is None:
            return query
        if isinstance(value, list) or isinstance(value, tuple):
            return query.filter(column.not_in(value))
        return query.filter(column != value)

    def getBuilds(self, tenant=None, project=None, pipeline=None,
                  change=None, branch=None, patchset=None, ref=None,
                  newrev=None, event_id=None, event_timestamp=None,
                  first_build_start_time=None, last_build_end_time=None,
                  queue_item_uuid=None, uuid=None, job_name=None,
                  voting=None, nodeset=None,
                  result=None, provides=None, final=None, held=None,
                  complete=None, sort_by_buildset=False, limit=50,
                  offset=0, idx_min=None, idx_max=None,
                  exclude_result=None, query_timeout=None):

        ref_table = self.connection.zuul_ref_table
        build_table = self.connection.zuul_build_table
        buildset_table = self.connection.zuul_buildset_table
        provides_table = self.connection.zuul_provides_table

        q = self.session().query(
            self.connection.buildModel.id).\
            join(self.connection.buildSetModel).\
            join(self.connection.refModel).\
            group_by(self.connection.buildModel.id)

        # If the query planner isn't able to reduce either the number
        # of rows returned by the buildset or build tables, then it
        # tends to produce a very slow query.  This hint produces
        # better results, but only in those cases.  When we can narrow
        # things down with indexes, it's better to omit the hint.
        # job_name is a tricky one.  It is indexed, but if there are a
        # lot of rows, it is better to include the hint, but if there
        # are few, it is better to not include it.  We include the hint
        # regardless of whether job_name is specified (optimizing for
        # the more common case).
        if not (change or uuid):
            q = q.with_hint(build_table, 'USE INDEX (PRIMARY)', 'mysql')
            q = q.with_hint(build_table, 'USE INDEX (PRIMARY)', 'mariadb')

        # Avoid joining the provides table unless necessary; postgres
        # has created some poor query plans in that case.  Currently
        # the only time this gets called with provides is from the
        # scheduler which has several other criteria which narrow down
        # the query.
        if provides:
            q = q.outerjoin(self.connection.providesModel)

        q = self.listFilter(q, buildset_table.c.tenant, tenant)
        q = self.listFilterFuzzy(q, buildset_table.c.pipeline, pipeline)
        q = self.listFilterFuzzy(q, ref_table.c.project, project)
        q = self.listFilter(q, ref_table.c.change, change)
        q = self.listFilterFuzzy(q, ref_table.c.branch, branch)
        q = self.listFilter(q, ref_table.c.patchset, patchset)
        q = self.listFilter(q, ref_table.c.ref, ref)
        q = self.listFilter(q, ref_table.c.newrev, newrev)
        q = self.listFilter(q, buildset_table.c.event_id, event_id)
        q = self.listFilter(
            q, buildset_table.c.event_timestamp, event_timestamp)
        q = self.listFilter(
            q, buildset_table.c.first_build_start_time, first_build_start_time)
        q = self.listFilter(
            q, buildset_table.c.last_build_end_time, last_build_end_time)
        q = self.listFilter(
            q, buildset_table.c.queue_item_uuid, queue_item_uuid)
        q = self.listFilter(q, build_table.c.uuid, uuid)
        q = self.listFilterFuzzy(q, build_table.c.job_name, job_name)
        q = self.listFilter(q, build_table.c.voting, voting)
        q = self.listFilter(q, build_table.c.nodeset, nodeset)
        q = self.listFilter(q, build_table.c.result, result)
        q = self.exListFilter(q, build_table.c.result, exclude_result)
        q = self.listFilter(q, build_table.c.final, final)
        if complete is True:
            q = q.filter(build_table.c.result != None)  # noqa
        elif complete is False:
            q = q.filter(build_table.c.result == None)  # noqa
        q = self.listFilter(q, provides_table.c.name, provides)
        q = self.listFilter(q, build_table.c.held, held)
        if idx_min:
            q = q.filter(build_table.c.id >= idx_min)
        if idx_max:
            q = q.filter(build_table.c.id <= idx_max)

        if sort_by_buildset:
            # If we don't need the builds to be strictly ordered, this
            # query can be much faster as it may avoid the use of a
            # temporary table.
            q = q.order_by(buildset_table.c.id.desc())
        else:
            q = q.order_by(build_table.c.id.desc())

        q = q.limit(limit).offset(offset)

        if query_timeout:
            # For MySQL, we can add a query hint directly.
            q = q.prefix_with(
                f'/*+ MAX_EXECUTION_TIME({query_timeout}) */',
                dialect='mysql')
            # For Postgres and mariadb, we add a comment that we parse
            # in our event handler.
            q = q.with_statement_hint(
                f'/* postgres_statement_timeout={query_timeout} */',
                dialect_name='postgresql')
            q = q.with_statement_hint(
                f'/* mariadb_statement_timeout={query_timeout} */',
                dialect_name='mariadb')

        ids = [x[0] for x in q.all()]
        # We use two queries here, principally because mariadb does
        # not produce a reasonable plan if we combine these into one
        # query with a join, and neither mysql nor mariadb support
        # using limit inside a subquery that is used with the "in"
        # operator.  Therefore, we run a query to get the ids, then
        # run a second to load the data.
        # contains_eager allows us to perform eager loading on the
        # buildset *and* use that table in filters (unlike
        # joinedload).
        bq = self.session().query(self.connection.buildModel).\
            join(self.connection.buildSetModel).\
            join(self.connection.refModel).\
            options(orm.contains_eager(self.connection.buildModel.buildset),
                    orm.contains_eager(self.connection.buildModel.ref),
                    orm.selectinload(self.connection.buildModel.provides),
                    orm.selectinload(self.connection.buildModel.artifacts))

        bq = bq.filter(self.connection.buildModel.id.in_(ids))

        if sort_by_buildset:
            # If we don't need the builds to be strictly ordered, this
            # query can be much faster as it may avoid the use of a
            # temporary table.
            bq = bq.order_by(buildset_table.c.id.desc())
        else:
            bq = bq.order_by(build_table.c.id.desc())

        try:
            return bq.all()
        except sqlalchemy.orm.exc.NoResultFound:
            return []

    def getBuild(self, tenant, uuid):
        build_table = self.connection.zuul_build_table
        buildset_table = self.connection.zuul_buildset_table

        # contains_eager allows us to perform eager loading on the
        # buildset *and* use that table in filters (unlike
        # joinedload).
        q = self.session().query(self.connection.buildModel).\
            join(self.connection.buildSetModel).\
            join(self.connection.refModel).\
            outerjoin(self.connection.providesModel).\
            options(orm.contains_eager(self.connection.buildModel.buildset).
                    subqueryload(self.connection.buildSetModel.refs),
                    orm.selectinload(self.connection.buildModel.provides),
                    orm.selectinload(self.connection.buildModel.artifacts),
                    orm.selectinload(self.connection.buildModel.ref))

        q = self.listFilter(q, buildset_table.c.tenant, tenant)
        q = self.listFilter(q, build_table.c.uuid, uuid)

        try:
            return q.one()
        except sqlalchemy.orm.exc.NoResultFound:
            return None
        except sqlalchemy.orm.exc.MultipleResultsFound:
            raise Exception("Multiple builds found with uuid %s", uuid)

    def getBuildTimes(self, tenant=None, project=None, pipeline=None,
                      branch=None,
                      ref=None,
                      start_time=None,
                      end_time=None,
                      job_name=None,
                      final=None,
                      sort_by_buildset=False,
                      limit=50,
                      offset=0,
                      result=None,
                      exclude_result=None,
                      query_timeout=None):

        ref_table = self.connection.zuul_ref_table
        build_table = self.connection.zuul_build_table
        buildset_table = self.connection.zuul_buildset_table

        q = self.session().query(
            self.connection.buildModel.id,
            self.connection.buildModel.end_time).\
            join(self.connection.buildSetModel).\
            join(self.connection.refModel).\
            group_by(self.connection.buildModel.id)

        # See note above about the hint.
        if not (project):
            q = q.with_hint(build_table, 'USE INDEX (PRIMARY)', 'mysql')
            q = q.with_hint(build_table, 'USE INDEX (PRIMARY)', 'mariadb')

        q = self.listFilter(q, buildset_table.c.tenant, tenant)
        q = self.listFilter(q, buildset_table.c.pipeline, pipeline)
        q = self.listFilter(q, ref_table.c.project, project)
        q = self.listFilter(q, ref_table.c.branch, branch)
        q = self.listFilter(q, ref_table.c.ref, ref)
        q = self.listFilter(q, build_table.c.job_name, job_name)
        q = self.listFilter(q, build_table.c.result, result)
        q = self.exListFilter(q, build_table.c.result, exclude_result)
        q = self.listFilter(q, build_table.c.final, final)
        if start_time:
            # Intentionally using end_time here
            q = q.filter(build_table.c.end_time >= start_time)
        if end_time:
            q = q.filter(build_table.c.end_time <= end_time)
        # Only complete builds (but if we specify a result, then it's
        # obviously complete and this extra filter is not necessary)
        if result is None:
            q = q.filter(build_table.c.result != None)  # noqa

        q = q.order_by(build_table.c.end_time.desc()).\
            limit(limit).\
            offset(offset)

        if query_timeout:
            # For MySQL, we can add a query hint directly.
            q = q.prefix_with(
                f'/*+ MAX_EXECUTION_TIME({query_timeout}) */',
                dialect='mysql')
            # For Postgres and mariadb, we add a comment that we parse
            # in our event handler.
            q = q.with_statement_hint(
                f'/* postgres_statement_timeout={query_timeout} */',
                dialect_name='postgresql')
            q = q.with_statement_hint(
                f'/* mariadb_statement_timeout={query_timeout} */',
                dialect_name='mariadb')

        ids = [x[0] for x in q.all()]
        # contains_eager allows us to perform eager loading on the
        # buildset *and* use that table in filters (unlike
        # joinedload).
        bq = self.session().query(self.connection.buildModel).\
            join(self.connection.buildSetModel).\
            join(self.connection.refModel).\
            options(orm.contains_eager(self.connection.buildModel.buildset),
                    orm.contains_eager(self.connection.buildModel.ref)).\
            filter(self.connection.buildModel.id.in_(ids)).\
            order_by(build_table.c.end_time.desc())

        try:
            return bq.all()
        except sqlalchemy.orm.exc.NoResultFound:
            return []

    def createBuildSet(self, *args, **kw):
        bs = self.connection.buildSetModel(*args, **kw)
        self.session().add(bs)
        self.session().flush()
        return bs

    def getOrCreateRef(self, project, ref, ref_url,
                       change=None, patchset=None, branch=None,
                       oldrev=None, newrev=None):
        ref_table = self.connection.zuul_ref_table
        q = self.session().query(self.connection.refModel)
        # We only query for the columns that we include in the unique
        # constraint (i.e., we omit branch and ref_url which should be
        # guaranteed to be unique by ref).
        q = q.filter(ref_table.c.project == project,
                     ref_table.c.ref == ref,
                     ref_table.c.change == change,
                     ref_table.c.patchset == patchset,
                     ref_table.c.oldrev == oldrev,
                     ref_table.c.newrev == newrev)
        ret = q.all()
        if ret:
            return ret[0]
        ret = self.connection.refModel(
            project=project, ref=ref, ref_url=ref_url,
            change=change, patchset=patchset, branch=branch,
            oldrev=oldrev, newrev=newrev)
        self.session().add(ret)
        self.session().flush()
        return ret

    def getBuildsets(self, tenant=None, project=None, pipeline=None,
                     change=None, branch=None, patchset=None, ref=None,
                     newrev=None, uuid=None, queue_item_uuid=None,
                     result=None, complete=None, updated_max=None,
                     limit=50, offset=0, idx_min=None, idx_max=None,
                     exclude_result=None, query_timeout=None):

        buildset_table = self.connection.zuul_buildset_table
        buildset_ref_table = self.connection.zuul_buildset_ref_table
        ref_table = self.connection.zuul_ref_table

        q = self.session().query(
            self.connection.buildSetModel.id).\
            join(self.connection.buildSetRefModel).\
            join(self.connection.refModel).\
            group_by(self.connection.buildSetModel.id)

        # See note above about the hint.
        if not (change or uuid):
            q = q.with_hint(buildset_table, 'USE INDEX (PRIMARY)', 'mysql')
            q = q.with_hint(buildset_table, 'USE INDEX (PRIMARY)', 'mariadb')
            q = q.with_hint(buildset_ref_table, 'USE INDEX (PRIMARY)',
                            'mysql')
            q = q.with_hint(buildset_ref_table, 'USE INDEX (PRIMARY)',
                            'mariadb')

        q = self.listFilter(q, buildset_table.c.tenant, tenant)
        q = self.listFilterFuzzy(q, buildset_table.c.pipeline, pipeline)
        q = self.listFilterFuzzy(q, ref_table.c.project, project)
        q = self.listFilter(q, ref_table.c.change, change)
        q = self.listFilterFuzzy(q, ref_table.c.branch, branch)
        q = self.listFilter(q, ref_table.c.patchset, patchset)
        q = self.listFilter(q, ref_table.c.ref, ref)
        q = self.listFilter(q, ref_table.c.newrev, newrev)
        q = self.listFilter(q, buildset_table.c.uuid, uuid)
        q = self.listFilter(
            q, buildset_table.c.queue_item_uuid, queue_item_uuid)
        q = self.listFilter(q, buildset_table.c.result, result)
        q = self.exListFilter(q, buildset_table.c.result, exclude_result)
        if idx_min:
            q = q.filter(buildset_table.c.id >= idx_min)
        if idx_max:
            q = q.filter(buildset_table.c.id <= idx_max)

        if complete is True:
            q = q.filter(buildset_table.c.result != None)  # noqa
        elif complete is False:
            q = q.filter(buildset_table.c.result == None)  # noqa

        if updated_max:
            q = q.filter(buildset_table.c.updated < updated_max)
        q = q.order_by(buildset_table.c.id.desc()).\
            limit(limit).\
            offset(offset)

        if query_timeout:
            # For MySQL, we can add a query hint directly.
            q = q.prefix_with(
                f'/*+ MAX_EXECUTION_TIME({query_timeout}) */',
                dialect='mysql')
            # For Postgres and mariadb, we add a comment that we parse
            # in our event handler.
            q = q.with_statement_hint(
                f'/* postgres_statement_timeout={query_timeout} */',
                dialect_name='postgresql')
            q = q.with_statement_hint(
                f'/* mariadb_statement_timeout={query_timeout} */',
                dialect_name='mariadb')

        ids = [x[0] for x in q.all()]
        bq = self.session().query(self.connection.buildSetModel).\
            join(self.connection.buildSetRefModel).\
            join(self.connection.refModel).\
            options(orm.contains_eager(self.connection.buildSetModel.refs)).\
            filter(self.connection.buildSetModel.id.in_(ids)).\
            order_by(buildset_table.c.id.desc())

        try:
            return bq.all()
        except sqlalchemy.orm.exc.NoResultFound:
            return []

    def getBuildset(self, tenant, uuid):
        """Get one buildset with its builds"""

        buildset_table = self.connection.zuul_buildset_table

        q = self.session().query(self.connection.buildSetModel).\
            options(orm.joinedload(self.connection.buildSetModel.refs)).\
            options(orm.joinedload(
                self.connection.buildSetModel.buildset_events)).\
            options(orm.joinedload(self.connection.buildSetModel.builds).
                    subqueryload(self.connection.buildModel.artifacts)).\
            options(orm.joinedload(self.connection.buildSetModel.builds).
                    subqueryload(self.connection.buildModel.provides)).\
            options(orm.joinedload(self.connection.buildSetModel.builds).
                    subqueryload(self.connection.buildModel.ref))

        q = self.listFilter(q, buildset_table.c.tenant, tenant)
        q = self.listFilter(q, buildset_table.c.uuid, uuid)

        try:
            return q.one()
        except sqlalchemy.orm.exc.NoResultFound:
            return None
        except sqlalchemy.orm.exc.MultipleResultsFound:
            raise Exception("Multiple buildset found with uuid %s", uuid)

    def _getBuildsetsForDelete(self, updated_max, limit):
        # This does a query for buildsets, then a "select in" query
        # for each adjoining table.  This is used by the
        # deleteBuildsets method.
        buildset_table = self.connection.zuul_buildset_table
        q = self.session().query(self.connection.buildSetModel).\
            options(
                orm.selectinload(
                    self.connection.buildSetModel.builds,
                ),
                orm.selectinload(
                    self.connection.buildSetModel.refs,
                ),
                orm.selectinload(
                    self.connection.buildSetModel.buildset_events,
                ),
                orm.selectinload(
                    self.connection.buildSetModel.builds,
                    self.connection.buildModel.provides,
                ),
                orm.selectinload(
                    self.connection.buildSetModel.builds,
                    self.connection.buildModel.artifacts,
                ),
                orm.selectinload(
                    self.connection.buildSetModel.builds,
                    self.connection.buildModel.build_events,
                ))

        q = q.filter(buildset_table.c.updated < updated_max)
        q = q.order_by(buildset_table.c.id.desc()).limit(limit)
        try:
            return q.all()
        except sqlalchemy.orm.exc.NoResultFound:
            return []

    def deleteBuildsets(self, cutoff, batch_size):
        """Delete buildsets before the cutoff"""

        deleted = True
        while deleted:
            deleted = False
            oldest = None
            for buildset in self._getBuildsetsForDelete(
                    updated_max=cutoff, limit=batch_size):
                deleted = True
                if oldest is None:
                    oldest = buildset.updated
                else:
                    oldest = min(oldest, buildset.updated)
                self.session().delete(buildset)
            self.session().commit()
            if deleted:
                self.log.info("Deleted from %s to %s", oldest, cutoff)

    def createSystemEvent(self, *args, **kw):
        """Create a new system event"""
        e = self.connection.systemEventModel(*args, **kw)
        self.session().add(e)
        self.session().flush()
        return e

    def getSystemEvents(self, tenant=None, event_type=None,
                        limit=50, offset=0, idx_min=None, idx_max=None,
                        query_timeout=None):

        event_table = self.connection.zuul_system_event_table

        q = self.session().query(self.connection.systemEventModel)
        q = self.listFilter(q, event_table.c.tenant, tenant)
        q = self.listFilter(q, event_table.c.event_type, event_type)
        if idx_min:
            q = q.filter(event_table.c.id >= idx_min)
        if idx_max:
            q = q.filter(event_table.c.id <= idx_max)

        q = q.order_by(event_table.c.id.desc()).\
            limit(limit).\
            offset(offset)

        if query_timeout:
            # For MySQL, we can add a query hint directly.
            q = q.prefix_with(
                f'/*+ MAX_EXECUTION_TIME({query_timeout}) */',
                dialect='mysql')
            # For Postgres and mariadb, we add a comment that we parse
            # in our event handler.
            q = q.with_statement_hint(
                f'/* postgres_statement_timeout={query_timeout} */',
                dialect_name='postgresql')
            q = q.with_statement_hint(
                f'/* mariadb_statement_timeout={query_timeout} */',
                dialect_name='mariadb')

        try:
            return q.all()
        except sqlalchemy.orm.exc.NoResultFound:
            return []

    def deleteSystemEvents(self, cutoff):
        """Delete system events before the cutoff"""

        event_table = self.connection.zuul_system_event_table
        q = sa.delete(self.connection.systemEventModel)\
              .where(event_table.c.event_time < cutoff)
        result = self.session().execute(q)
        self.session().commit()
        if result.rowcount:
            self.log.info("Deleted %s events until %s",
                          result.rowcount, cutoff)


class SQLConnection(BaseConnection):
    driver_name = 'sql'
    log = logging.getLogger("zuul.SQLConnection")
    # This is used by tests only
    force_migrations = False

    def __init__(self, driver, connection_name, connection_config):

        super(SQLConnection, self).__init__(driver, connection_name,
                                            connection_config)

        self.dburi = None
        self.engine = None
        self.connection = None
        self.table_prefix = self.connection_config.get('table_prefix', '')
        self.log.info("Initializing SQL connection {} (prefix: {})".format(
            connection_name, self.table_prefix))

        try:
            self.dburi = self.connection_config.get('dburi')
            self.metadata = sa.MetaData()

            # Recycle connections if they've been idle for more than 1 second.
            # MySQL connections are lightweight and thus keeping long-lived
            # connections around is not valuable.
            self.engine = sa.create_engine(
                self.dburi,
                poolclass=sqlalchemy.pool.QueuePool,
                pool_recycle=self.connection_config.get('pool_recycle', 1),
                future=True)

            self._setup_models()

            # If we want the objects returned from query() to be
            # usable outside of the session, we need to expunge them
            # from the session, and since the DatabaseSession always
            # calls commit() on the session when the context manager
            # exits, we need to inform the session not to expire
            # objects when it does so.
            self.session_factory = orm.sessionmaker(bind=self.engine,
                                                    expire_on_commit=False,
                                                    autoflush=False,
                                                    future=True)
            self.session = orm.scoped_session(self.session_factory)
        except sa.exc.NoSuchModuleError:
            self.log.error(
                "The required module for the dburi dialect isn't available.")
            raise

    def getSession(self):
        return DatabaseSession(self)

    def _migrate(self, revision='head'):
        """Perform the alembic migrations for this connection"""
        # Note that this method needs to be called with an external lock held.
        # The reason for this is we retrieve the alembic version and run the
        # alembic migrations in different database transactions which opens
        # us to races without an external lock.
        with self.engine.begin() as conn:
            context = alembic.migration.MigrationContext.configure(conn)
            current_rev = context.get_current_revision()
        self.log.debug('Current migration revision: %s' % current_rev)

        config = alembic.config.Config()
        config.set_main_option("script_location",
                               "zuul:driver/sql/alembic")
        config.set_main_option("sqlalchemy.url",
                               self.connection_config.get('dburi').
                               replace('%', '%%'))

        # Alembic lets us add arbitrary data in the tag argument. We can
        # leverage that to tell the upgrade scripts about the table prefix.
        tag = {'table_prefix': self.table_prefix}

        if current_rev is None and not self.force_migrations:
            self.metadata.create_all(self.engine)
            alembic.command.stamp(config, revision, tag=tag)
        else:
            alembic.command.upgrade(config, revision, tag=tag)

    def onLoad(self, zk_client, component_registry=None):
        safe_connection = quote_plus(self.connection_name)
        while True:
            try:
                with locked(
                    SessionAwareLock(
                        zk_client.client,
                        f"{CONNECTION_LOCK_ROOT}/{safe_connection}/migration")
                ):
                    self._migrate()
                break
            except sa.exc.OperationalError as e:
                self.log.error(
                    "Unable to connect to the database or establish the "
                    "required tables: %s", e)
            except Exception:
                self.log.exception("Error setting up database:")
                raise
            time.sleep(10)

    def _setup_models(self):
        Base = orm.declarative_base(metadata=self.metadata)

        class RefModel(Base):
            __tablename__ = self.table_prefix + REF_TABLE
            id = sa.Column(sa.Integer, primary_key=True)
            project = sa.Column(sa.String(SQL_MAX_STRING_LENGTH),
                                nullable=False)
            ref = sa.Column(sa.String(SQL_MAX_STRING_LENGTH),
                            nullable=False)
            ref_url = sa.Column(sa.String(SQL_MAX_STRING_LENGTH),
                                nullable=False)
            change = sa.Column(ChangeType, nullable=False)
            patchset = sa.Column(SHAType(40), nullable=False)
            oldrev = sa.Column(SHAType(40), nullable=False)
            newrev = sa.Column(SHAType(40), nullable=False)
            branch = sa.Column(sa.String(SQL_MAX_STRING_LENGTH),
                               nullable=False)

            sa.Index(self.table_prefix + 'zuul_ref_project_idx', project)
            sa.Index(self.table_prefix + 'zuul_ref_ref_idx', ref)
            sa.Index(self.table_prefix + 'zuul_ref_change_idx', change)
            sa.Index(self.table_prefix + 'zuul_ref_patchset_idx', patchset)
            sa.Index(self.table_prefix + 'zuul_ref_oldrev_idx', oldrev)
            sa.Index(self.table_prefix + 'zuul_ref_newrev_idx', newrev)
            sa.Index(self.table_prefix + 'zuul_ref_branch_idx', branch)
            sa.UniqueConstraint(
                project, ref, change, patchset, oldrev, newrev,
                name=self.table_prefix + 'zuul_ref_unique')

        class BuildSetModel(Base):
            __tablename__ = self.table_prefix + BUILDSET_TABLE
            id = sa.Column(sa.Integer, primary_key=True)
            pipeline = sa.Column(sa.String(SQL_MAX_STRING_LENGTH))
            message = sa.Column(sa.TEXT())
            tenant = sa.Column(sa.String(SQL_MAX_STRING_LENGTH))
            result = sa.Column(sa.String(SQL_MAX_STRING_LENGTH))
            uuid = sa.Column(sa.String(36))
            event_id = sa.Column(sa.String(SQL_MAX_STRING_LENGTH),
                                 nullable=True)
            event_timestamp = sa.Column(sa.DateTime, nullable=True)
            first_build_start_time = sa.Column(sa.DateTime, nullable=True)
            last_build_end_time = sa.Column(sa.DateTime, nullable=True)
            updated = sa.Column(sa.DateTime, nullable=True)
            queue_item_uuid = sa.Column(sa.String(36), nullable=True)

            refs = orm.relationship(
                RefModel,
                secondary=self.table_prefix + BUILDSET_REF_TABLE)
            sa.Index(self.table_prefix + 'zuul_buildset_uuid_idx', uuid)
            sa.Index(self.table_prefix + 'zuul_buildset_tenant_idx', tenant)
            sa.Index(self.table_prefix + 'zuul_buildset_queue_item_uuid_idx',
                     queue_item_uuid)

            def createBuild(self, ref, *args, **kw):
                session = orm.session.Session.object_session(self)
                b = BuildModel(*args, **kw)
                b.buildset_id = self.id
                b.ref_id = ref.id
                self.builds.append(b)
                session.add(b)
                session.flush()
                return b

            def createBuildSetEvent(self, *args, **kw):
                session = orm.session.Session.object_session(self)
                e = BuildSetEventModel(*args, **kw)
                e.buildset_id = self.id
                self.buildset_events.append(e)
                session.add(e)
                session.flush()
                return e

        class BuildSetRefModel(Base):
            __tablename__ = self.table_prefix + BUILDSET_REF_TABLE
            __table_args__ = (
                sa.PrimaryKeyConstraint('buildset_id', 'ref_id'),
            )
            buildset_id = sa.Column(sa.Integer, sa.ForeignKey(
                self.table_prefix + BUILDSET_TABLE + ".id",
                name=self.table_prefix + 'zuul_buildset_ref_buildset_id_fkey',
            ))
            ref_id = sa.Column(sa.Integer, sa.ForeignKey(
                self.table_prefix + REF_TABLE + ".id",
                name=self.table_prefix + 'zuul_buildset_ref_ref_id_fkey',
            ))
            sa.Index(self.table_prefix + 'zuul_buildset_ref_buildset_id_idx',
                     buildset_id)
            sa.Index(self.table_prefix + 'zuul_buildset_ref_ref_id_idx',
                     ref_id)

        class BuildSetEventModel(Base):
            __tablename__ = self.table_prefix + BUILDSET_EVENT_TABLE
            id = sa.Column(sa.Integer, primary_key=True)
            buildset_id = sa.Column(sa.Integer, sa.ForeignKey(
                self.table_prefix + BUILDSET_TABLE + ".id",
                name=(self.table_prefix +
                      'zuul_buildset_event_buildset_id_fkey'),
            ))
            event_time = sa.Column(sa.DateTime)
            event_type = sa.Column(sa.String(SQL_MAX_STRING_LENGTH))
            description = sa.Column(sa.TEXT())
            buildset = orm.relationship(BuildSetModel,
                                        backref=orm.backref(
                                            "buildset_events",
                                            cascade="all, delete-orphan"))
            sa.Index(self.table_prefix + 'zuul_buildset_event_buildset_id_idx',
                     buildset_id)

        class BuildModel(Base):
            __tablename__ = self.table_prefix + BUILD_TABLE
            id = sa.Column(sa.Integer, primary_key=True)
            buildset_id = sa.Column(sa.Integer, sa.ForeignKey(
                self.table_prefix + BUILDSET_TABLE + ".id",
                name=self.table_prefix + 'zuul_build_buildset_id_fkey',
            ))
            uuid = sa.Column(sa.String(36))
            job_name = sa.Column(sa.String(SQL_MAX_STRING_LENGTH))
            result = sa.Column(sa.String(SQL_MAX_STRING_LENGTH))
            start_time = sa.Column(sa.DateTime)
            end_time = sa.Column(sa.DateTime)
            voting = sa.Column(sa.Boolean)
            log_url = sa.Column(sa.String(SQL_MAX_STRING_LENGTH))
            error_detail = sa.Column(sa.TEXT())
            final = sa.Column(sa.Boolean)
            held = sa.Column(sa.Boolean)
            nodeset = sa.Column(sa.String(SQL_MAX_STRING_LENGTH))
            ref_id = sa.Column(sa.Integer, sa.ForeignKey(
                self.table_prefix + REF_TABLE + ".id",
                name=self.table_prefix + 'zuul_build_ref_id_fkey',
            ))
            execute_time = sa.Column(sa.DateTime)

            buildset = orm.relationship(BuildSetModel,
                                        backref=orm.backref(
                                            "builds",
                                            order_by=id,
                                            cascade="all, delete-orphan"))
            ref = orm.relationship(RefModel)

            sa.Index(self.table_prefix + 'zuul_build_job_name_buildset_id_idx',
                     job_name, buildset_id)
            sa.Index(self.table_prefix + 'zuul_build_uuid_buildset_id_idx',
                     uuid, buildset_id)
            sa.Index(self.table_prefix + 'zuul_build_buildset_id_idx',
                     buildset_id)
            sa.Index(self.table_prefix + 'zuul_build_ref_id_idx',
                     ref_id)
            sa.Index(self.table_prefix + 'zuul_build_end_time_idx',
                     end_time)

            @property
            def duration(self):
                if self.start_time and self.end_time:
                    return max(0.0,
                               (self.end_time -
                                self.start_time).total_seconds())
                else:
                    return None

            def createArtifact(self, *args, **kw):
                session = orm.session.Session.object_session(self)
                # SQLAlchemy reserves the 'metadata' attribute on
                # object models, so our model and table names use
                # 'meta', but here we accept data directly from
                # zuul_return where it's called 'metadata'.  Transform
                # the attribute name.
                if 'metadata' in kw:
                    kw['meta'] = kw['metadata']
                    del kw['metadata']
                a = ArtifactModel(*args, **kw)
                a.build_id = self.id
                self.artifacts.append(a)
                session.add(a)
                session.flush()
                return a

            def createProvides(self, *args, **kw):
                session = orm.session.Session.object_session(self)
                p = ProvidesModel(*args, **kw)
                p.build_id = self.id
                self.provides.append(p)
                session.add(p)
                session.flush()
                return p

            def createBuildEvent(self, *args, **kw):
                session = orm.session.Session.object_session(self)
                e = BuildEventModel(*args, **kw)
                e.build_id = self.id
                self.build_events.append(e)
                session.add(e)
                session.flush()
                return e

        class ArtifactModel(Base):
            __tablename__ = self.table_prefix + ARTIFACT_TABLE
            id = sa.Column(sa.Integer, primary_key=True)
            build_id = sa.Column(sa.Integer, sa.ForeignKey(
                self.table_prefix + BUILD_TABLE + ".id",
                name=self.table_prefix + 'zuul_artifact_build_id_fkey',
            ))
            name = sa.Column(sa.String(SQL_MAX_STRING_LENGTH))
            url = sa.Column(sa.TEXT())
            meta = sa.Column('metadata', sa.TEXT())
            build = orm.relationship(BuildModel,
                                     backref=orm.backref(
                                         "artifacts",
                                         cascade="all, delete-orphan"))
            sa.Index(self.table_prefix + 'zuul_artifact_build_id_idx',
                     build_id)

        class ProvidesModel(Base):
            __tablename__ = self.table_prefix + PROVIDES_TABLE
            id = sa.Column(sa.Integer, primary_key=True)
            build_id = sa.Column(sa.Integer, sa.ForeignKey(
                self.table_prefix + BUILD_TABLE + ".id",
                name=self.table_prefix + 'zuul_provides_build_id_fkey',
            ))
            name = sa.Column(sa.String(SQL_MAX_STRING_LENGTH))
            build = orm.relationship(BuildModel,
                                     backref=orm.backref(
                                         "provides",
                                         cascade="all, delete-orphan"))
            sa.Index(self.table_prefix + 'zuul_provides_build_id_idx',
                     build_id)

        class BuildEventModel(Base):
            __tablename__ = self.table_prefix + BUILD_EVENT_TABLE
            id = sa.Column(sa.Integer, primary_key=True)
            build_id = sa.Column(sa.Integer, sa.ForeignKey(
                self.table_prefix + BUILD_TABLE + ".id",
                name=self.table_prefix + 'zuul_build_event_build_id_fkey',
            ))
            event_time = sa.Column(sa.DateTime)
            event_type = sa.Column(sa.String(SQL_MAX_STRING_LENGTH))
            description = sa.Column(sa.TEXT())
            build = orm.relationship(BuildModel,
                                     backref=orm.backref(
                                         "build_events",
                                         cascade="all, delete-orphan"))
            sa.Index(self.table_prefix + 'zuul_build_event_build_id_idx',
                     build_id)

        class SystemEventModel(Base):
            __tablename__ = self.table_prefix + SYSTEM_EVENT_TABLE
            id = sa.Column(sa.Integer, primary_key=True)
            tenant = sa.Column(sa.String(SQL_MAX_STRING_LENGTH))
            event_id = sa.Column(sa.String(SQL_MAX_STRING_LENGTH))
            event_time = sa.Column(sa.DateTime)
            event_type = sa.Column(sa.String(SQL_MAX_STRING_LENGTH))
            description = sa.Column(sa.TEXT())

            sa.Index(self.table_prefix + 'zuul_system_event_tenant_idx',
                     tenant)
            sa.Index(self.table_prefix + 'zuul_system_event_event_id_idx',
                     event_id)
            sa.Index(self.table_prefix + 'zuul_system_event_event_time_idx',
                     event_time)
            sa.Index(self.table_prefix + 'zuul_system_event_event_type_idx',
                     event_type)

        self.buildEventModel = BuildEventModel
        self.zuul_build_event_table = self.buildEventModel.__table__

        self.buildSetEventModel = BuildSetEventModel
        self.zuul_buildset_event_table = self.buildSetEventModel.__table__

        self.providesModel = ProvidesModel
        self.zuul_provides_table = self.providesModel.__table__

        self.artifactModel = ArtifactModel
        self.zuul_artifact_table = self.artifactModel.__table__

        self.buildModel = BuildModel
        self.zuul_build_table = self.buildModel.__table__

        self.buildSetModel = BuildSetModel
        self.zuul_buildset_table = self.buildSetModel.__table__

        self.refModel = RefModel
        self.zuul_ref_table = self.refModel.__table__

        self.buildSetRefModel = BuildSetRefModel
        self.zuul_buildset_ref_table = self.buildSetRefModel.__table__

        self.systemEventModel = SystemEventModel
        self.zuul_system_event_table = self.systemEventModel.__table__

    def onStop(self):
        self.log.debug("Stopping SQL connection %s" % self.connection_name)
        self.engine.dispose()

    def getBuilds(self, *args, **kw):
        """Return a list of Build objects"""
        with self.getSession() as db:
            return db.getBuilds(*args, **kw)

    def getBuild(self, *args, **kw):
        """Return a Build object"""
        with self.getSession() as db:
            return db.getBuild(*args, **kw)

    def getBuildsets(self, *args, **kw):
        """Return a list of BuildSet objects"""
        with self.getSession() as db:
            return db.getBuildsets(*args, **kw)

    def getBuildset(self, *args, **kw):
        """Return a BuildSet objects"""
        with self.getSession() as db:
            return db.getBuildset(*args, **kw)

    def deleteBuildsets(self, *args, **kw):
        """Delete buildsets"""
        with self.getSession() as db:
            return db.deleteBuildsets(*args, **kw)

    def getBuildTimes(self, *args, **kw):
        """Return a list of Build objects"""
        with self.getSession() as db:
            return db.getBuildTimes(*args, **kw)

    def getSystemEvents(self, *args, **kw):
        """Return a list of SystemEvent objects"""
        with self.getSession() as db:
            return db.getSystemEvents(*args, **kw)

    def deleteSystemEvents(self, *args, **kw):
        """Delete system events"""
        with self.getSession() as db:
            return db.deleteSystemEvents(*args, **kw)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import logging
import time
from abc import ABCMeta
from configparser import ConfigParser
from threading import Thread
from typing import List, Callable

import kazoo.client
from kazoo.exceptions import NoNodeError, NodeExistsError
from kazoo.handlers.threading import KazooTimeoutError
from kazoo.protocol.states import KazooState

from zuul.lib.config import get_default
from zuul.zk.exceptions import NoClientException
from zuul.zk.handler import PoolSequentialThreadingHandler
from zuul.zk.vendor.client import ZuulKazooClient
from zuul.zk.vendor.connection import ZuulConnectionHandler

kazoo.client.ConnectionHandler = ZuulConnectionHandler


class ZooKeeperClient(object):
    log = logging.getLogger("zuul.zk.base.ZooKeeperClient")

    # Log zookeeper retry every 10 seconds
    retry_log_rate = 10

    def __init__(
        self,
        hosts,
        read_only=False,
        timeout=10.0,
        tls_cert=None,
        tls_key=None,
        tls_ca=None,
    ):
        """
        Initialize the ZooKeeper base client object.

        :param str hosts: Comma-separated list of hosts to connect to (e.g.
            127.0.0.1:2181,127.0.0.1:2182,[::1]:2183).
        :param bool read_only: If True, establishes a read-only connection.
        :param float timeout: The ZooKeeper session timeout, in
            seconds (default: 10.0).
        :param str tls_key: Path to TLS key
        :param str tls_cert: Path to TLS cert
        :param str tls_ca: Path to TLS CA cert
        """
        self.hosts = hosts
        self.read_only = read_only
        self.timeout = timeout
        self.tls_cert = tls_cert
        self.tls_key = tls_key
        self.tls_ca = tls_ca
        self.was_lost = False
        self.cap_watched_event_zxid = False

        self.client = None

        # Verify that we can read the cert files (Kazoo doesn't
        # provide useful error messages).
        for fn in (tls_cert, tls_key, tls_ca):
            if fn:
                with open(fn):
                    pass

        self._last_retry_log: int = 0
        self.on_connect_listeners: List[Callable[[], None]] = []
        self.on_disconnect_listeners: List[Callable[[], None]] = []
        self.on_connection_lost_listeners: List[Callable[[], None]] = []
        self.on_reconnect_listeners: List[Callable[[], None]] = []
        self.on_suspended_listeners = []

    def _connectionListener(self, state):
        """
        Listener method for Kazoo connection state changes.

        .. warning:: This method must not block.
        """
        session_id = 0
        if client_id := self.client.client_id:
            session_id, _ = client_id

        if state == KazooState.LOST:
            self.log.debug("ZooKeeper connection (session: %#x): LOST",
                           session_id)
            self.was_lost = True
            for listener in self.on_connection_lost_listeners:
                try:
                    listener()
                except Exception:
                    self.log.exception("Exception calling listener:")
        elif state == KazooState.SUSPENDED:
            self.log.debug("ZooKeeper connection (session: %#x): SUSPENDED",
                           session_id)
            for listener in self.on_suspended_listeners:
                try:
                    listener()
                except Exception:
                    self.log.exception("Exception calling listener:")
        else:
            self.log.debug("ZooKeeper connection (session: %#x): CONNECTED",
                           session_id)
            # Create a throwaway thread since zk operations can't
            # happen in this one.
            if self.was_lost:
                self.was_lost = False
                for listener in self.on_reconnect_listeners:
                    t = Thread(target=listener)
                    t.daemon = True
                    t.start()

    @property
    def connected(self):
        return self.client and self.client.state == KazooState.CONNECTED

    @property
    def suspended(self):
        return self.client and self.client.state == KazooState.SUSPENDED

    @property
    def lost(self):
        return not self.client or self.client.state == KazooState.LOST

    def logConnectionRetryEvent(self):
        now = time.monotonic()
        if now - self._last_retry_log >= self.retry_log_rate:
            self.log.warning("Retrying zookeeper connection")
            self._last_retry_log = now

    def connect(self):
        if self.client is None:
            args = dict(
                hosts=self.hosts,
                read_only=self.read_only,
                timeout=self.timeout,
                handler=PoolSequentialThreadingHandler(),
            )
            if self.tls_key:
                args['use_ssl'] = True
                args['keyfile'] = self.tls_key
                args['certfile'] = self.tls_cert
                args['ca'] = self.tls_ca
            self.client = ZuulKazooClient(**args)
            self.client.add_listener(self._connectionListener)
            # Manually retry initial connection attempt
            while True:
                try:
                    self.client.start(1)
                    break
                except KazooTimeoutError:
                    self.logConnectionRetryEvent()

            self._checkCapabilities()
            for listener in self.on_connect_listeners:
                listener()

    def _checkCapabilities(self):
        self.client.ensure_path('/zuul')

        def watcher(event):
            if event.zxid != event.NO_ZXID:
                self.cap_watched_event_zxid = True

        # Create and delete a node with a watch to experimentally
        # determine whether we are running ZK 3.9+ with the zxid on
        # event feature.
        node = self.client.create('/zuul/capcheck',
                                  sequence=True, ephemeral=True)
        self.client.get(node, watch=watcher)
        self.client.delete(node)

    def disconnect(self):
        """
        Close the ZooKeeper cluster connection.

        You should call this method if you used connect() to establish a
        cluster connection.
        """
        for listener in self.on_disconnect_listeners:
            listener()

        if self.client is not None and self.client.connected:
            self.client.stop()
            self.client.close()
            self.client = None

    def resetHosts(self, hosts):
        """
        Reset the ZooKeeper cluster connection host list.

        :param str hosts: Comma-separated list of hosts to connect to (e.g.
            127.0.0.1:2181,127.0.0.1:2182,[::1]:2183).
        """
        if self.client is not None:
            self.client.set_hosts(hosts=hosts)

    @classmethod
    def fromConfig(cls, config: ConfigParser) -> "ZooKeeperClient":
        hosts = get_default(config, "zookeeper", "hosts")
        if not hosts:
            raise Exception("The zookeeper hosts config value is required")
        tls_key = get_default(config, "zookeeper", "tls_key")
        tls_cert = get_default(config, "zookeeper", "tls_cert")
        tls_ca = get_default(config, "zookeeper", "tls_ca")
        if not all([tls_key, tls_cert, tls_ca]):
            raise Exception(
                "A TLS ZooKeeper connection is required; please supply the "
                "tls_* zookeeper config values."
            )
        timeout = float(
            get_default(config, "zookeeper", "session_timeout", 120.0)
        )

        return cls(
            hosts=hosts,
            timeout=timeout,
            tls_key=tls_key,
            tls_cert=tls_cert,
            tls_ca=tls_ca,
        )

    def commitTransaction(self, tr):
        results = tr.commit()
        for res in results:
            self.log.debug("Transaction response %s", repr(res))
        for res in results:
            if isinstance(res, Exception):
                raise res
        return results

    def getCurrentLtime(self):
        """Get the logical timestamp as seen by the Zookeeper cluster."""
        try:
            zstat = self.client.set("/zuul/ltime", b"")
        except NoNodeError:
            try:
                self.client.create("/zuul/ltime", b"", makepath=True)
                zstat = self.client.set("/zuul/ltime", b"")
            except NodeExistsError:
                zstat = self.client.set("/zuul/ltime", b"")
        return zstat.last_modified_transaction_id

    def _fastRecursiveDelete(self, path, results):
        try:
            children = self.client.get_children(path)
        except NoNodeError:
            return
        if children:
            for child in children:
                self._fastRecursiveDelete(f'{path}/{child}', results)
        results.append(self.client.delete_async(path, -1))

    def fastRecursiveDelete(self, path):
        results = []
        self._fastRecursiveDelete(path, results)
        for res in results:
            try:
                res.get()
            except NoNodeError:
                pass


class ZooKeeperSimpleBase(metaclass=ABCMeta):
    """Base class for stateless Zookeeper interaction."""

    def __init__(self, client: ZooKeeperClient):
        self.client = client

    @property
    def kazoo_client(self) -> ZuulKazooClient:
        if not self.client.client:
            raise NoClientException()
        return self.client.client


class ZooKeeperBase(ZooKeeperSimpleBase):
    """Base class for registering state handling methods with ZooKeeper."""

    def __init__(self, client: ZooKeeperClient):
        super().__init__(client)
        if client:
            self.client.on_connect_listeners.append(self._onConnect)
            self.client.on_disconnect_listeners.append(self._onDisconnect)
            self.client.on_reconnect_listeners.append(self._onReconnect)
            self.client.on_suspended_listeners.append(self._onSuspended)

    def stop(self):
        if self.client:
            self.client.on_connect_listeners.remove(self._onConnect)
            self.client.on_disconnect_listeners.remove(self._onDisconnect)
            self.client.on_reconnect_listeners.remove(self._onReconnect)
            self.client.on_suspended_listeners.remove(self._onSuspended)

    def _onConnect(self):
        pass

    def _onDisconnect(self):
        pass

    def _onReconnect(self):
        pass

    def _onSuspended(self):
        pass



# Copyright 2012 Hewlett-Packard Development Company, L.P.
# Copyright 2013 OpenStack Foundation
# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import psutil

from zuul.driver import (Driver, WrapperInterface)
from zuul.execution_context import BaseExecutionContext


class NullExecutionContext(BaseExecutionContext):
    log = logging.getLogger("zuul.NullExecutionContext")

    def getPopen(self, **kwargs):
        return psutil.Popen

    def getNamespacePids(self, proc):
        return None, []


class NullwrapDriver(Driver, WrapperInterface):
    name = 'nullwrap'
    log = logging.getLogger("zuul.NullwrapDriver")

    def getExecutionContext(self, ro_paths=None, rw_paths=None, secrets=None):
        # The bubblewrap driver writes secrets to a tmpfs so that they
        # don't hit the disk (unless the kernel swaps the memory to
        # disk, which can be mitigated with encrypted swap).  We
        # haven't implemented similar functionality in nullwrap, so
        # for safety, raise an exception in that case.  If you are
        # interested in implementing this functionality, please
        # contact us on the mailing list.
        if secrets:
            raise NotImplementedError(
                "The nullwrap driver does not support the use of secrets. "
                "Consider using the bubblewrap driver instead.")
        return NullExecutionContext()



# Copyright (C) 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import patch

import testtools

from zuul import exceptions
from zuul.model import ProviderNode
from zuul.launcher.server import NodescanWorker, NodescanRequest
from zuul.zk import zkobject

from tests.base import (
    BaseTestCase,
    iterate_timeout,
    okay_tracebacks,
)
from tests.fake_nodescan import (
    FakeSocket,
    FakePoll,
    FakeTransport,
    NodescanFixture,
)


class DummyProviderNode(ProviderNode, subclass_id="dummy-nodescan"):
    pass


class TestNodescanWorker(BaseTestCase):

    def createZKContext(self, lock=None):
        return zkobject.ZKContext(self.zk_client, lock,
                                  None, self.log)

    def test_nodescan(self):
        # Test the nodescan worker
        self.useFixture(NodescanFixture())
        worker = NodescanWorker()
        node = DummyProviderNode()
        node._set(
            interface_ip='198.51.100.1',
            connection_port=22,
            connection_type='ssh',
            host_key_checking=True,
            boot_timeout=300,
        )
        worker.start()
        request = NodescanRequest(node, self.log)
        worker.addRequest(request)
        for _ in iterate_timeout(30, 'waiting for nodescan'):
            if request.complete:
                break
        result = request.result()
        self.assertEqual(result, ['fake key fake base64'])
        worker.stop()
        worker.join()

    def test_nodescan_connection_timeout(self):
        # Test a timeout during socket connection
        self.useFixture(NodescanFixture(poll_fail=True))
        worker = NodescanWorker()
        node = DummyProviderNode()
        node._set(
            interface_ip='198.51.100.1',
            connection_port=22,
            connection_type='ssh',
            host_key_checking=True,
            boot_timeout=1,
        )
        worker.start()
        request = NodescanRequest(node, self.log)
        worker.addRequest(request)
        for _ in iterate_timeout(30, 'waiting for nodescan'):
            if request.complete:
                break
        with testtools.ExpectedException(
                exceptions.ConnectionTimeoutException):
            request.result()
        worker.stop()
        worker.join()

    def test_nodescan_ssh_timeout(self):
        # Test a timeout during ssh connection
        self.useFixture(NodescanFixture(transport_fail=True))
        worker = NodescanWorker()
        node = DummyProviderNode()
        node._set(
            interface_ip='198.51.100.1',
            connection_port=22,
            connection_type='ssh',
            host_key_checking=True,
            boot_timeout=1,
        )
        worker.start()
        request = NodescanRequest(node, self.log)
        worker.addRequest(request)
        for _ in iterate_timeout(30, 'waiting for nodescan'):
            if request.complete:
                break
        with testtools.ExpectedException(
                exceptions.ConnectionTimeoutException):
            request.result()
        worker.stop()
        worker.join()

    @okay_tracebacks('Fake ssh error')
    def test_nodescan_ssh_error(self):
        # Test an ssh error
        self.useFixture(NodescanFixture(transport_active=False))
        worker = NodescanWorker()
        node = DummyProviderNode()
        node._set(
            interface_ip='198.51.100.1',
            connection_port=22,
            connection_type='ssh',
            host_key_checking=True,
            boot_timeout=1,
        )
        worker.start()
        request = NodescanRequest(node, self.log)
        worker.addRequest(request)
        for _ in iterate_timeout(30, 'waiting for nodescan'):
            if request.complete:
                break
        with testtools.ExpectedException(
                exceptions.ConnectionTimeoutException):
            request.result()
        worker.stop()
        worker.join()

    @patch('paramiko.transport.Transport')
    @patch('zuul.launcher.server.NodescanRequest._socket_class')
    @patch('zuul.launcher.server.NodescanWorker._poll_class')
    def test_nodescan_queue(self, mock_epoll, mock_socket, mock_transport):
        # Test the max_requests queing function
        fake_socket1 = FakeSocket()
        fake_socket2 = FakeSocket()
        fake_socket2.fd = 2
        # We get two sockets for each host
        sockets = [fake_socket1, fake_socket1, fake_socket2, fake_socket2]

        def getsocket(*args, **kw):
            return sockets.pop(0)

        mock_socket.side_effect = getsocket
        mock_epoll.return_value = FakePoll()
        mock_transport.return_value = FakeTransport()
        worker = NodescanWorker()
        worker.MAX_REQUESTS = 1
        node1 = DummyProviderNode()
        node1._set(
            interface_ip='198.51.100.1',
            connection_port=22,
            connection_type='ssh',
            host_key_checking=True,
            boot_timeout=300,
        )
        node2 = DummyProviderNode()
        node2._set(
            interface_ip='198.51.100.2',
            connection_port=22,
            connection_type='ssh',
            host_key_checking=True,
            boot_timeout=300,
        )

        request1 = NodescanRequest(node1, self.log)
        request2 = NodescanRequest(node2, self.log)
        worker.addRequest(request1)
        worker.addRequest(request2)
        worker.start()
        for _ in iterate_timeout(5, 'waiting for nodescan'):
            if request1.complete and request2.complete:
                break
        result1 = request1.result()
        result2 = request2.result()
        self.assertEqual(result1, ['fake key fake base64'])
        self.assertEqual(result2, ['fake key fake base64'])
        worker.stop()
        worker.join()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""buildset_updated

Revision ID: c7467b642498
Revises: 4647def24b32
Create Date: 2022-05-28 16:21:50.035877

"""

# revision identifiers, used by Alembic.
revision = 'c7467b642498'
down_revision = '4647def24b32'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade(table_prefix=''):
    op.add_column(
        table_prefix + "zuul_buildset", sa.Column('updated', sa.DateTime))

    connection = op.get_bind()
    connection.execute(
        sa.text(
            """
            UPDATE {buildset_table}
              SET updated=greatest(
                coalesce(first_build_start_time, '1970-01-01 00:00:00'),
                coalesce(last_build_end_time, '1970-01-01 00:00:00'),
                coalesce(event_timestamp, '1970-01-01 00:00:00'))
            """.format(buildset_table=table_prefix + "zuul_buildset")
        )
    )


def downgrade():
    raise Exception("Downgrades not supported")



# vim: tabstop=4 shiftwidth=4 softtabstop=4

#    Copyright 2011 OpenStack LLC
#    Copyright 2012 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

from importlib import metadata as importlib_metadata

zuul_distribution = importlib_metadata.distribution('zuul')
release_string = zuul_distribution.version

is_release = None
git_version = None
try:
    _metadata = json.loads(zuul_distribution.read_text('pbr.json'))
    if _metadata:
        is_release = _metadata['is_release']
        git_version = _metadata['git_version']
except Exception:
    pass


def get_version_string():
    if is_release:
        return release_string
    return "{} {}".format(release_string, git_version)
